%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Bousi-Prolog parser

:- module(parser, [
    parse_program/7,        % +ProgramFile, +OntologyFile, -Directives,
                            %   -Rules, -Equations, -LingTerms,
                            %   -Messages
    parse_query/6           % +ProgramPrefix, +String, -Query,
                            %   -NewHypoRules, -LingTerms, -Messages
   ]).

:- use_module(directives).
:- use_module(utilities).
:- use_module(dcg).

:- use_module(library(lists)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- set_prolog_flag(double_quotes, codes).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parser predicates for Bousi-Prolog files
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% parse_program(+ProgramFile, +OntologyFile, -Directives, -Rules, -Equations, -LingTerms, -Messages)
%
%     Reads a Bousi-Prolog program together with a Bousi-Prolog
%     ontology, performs a lexical, syntactic and semantic analysis,
%     and returns the following:
%
%      * All directives, rules (including facts) and equations found in
%        source code files.
%      * A list with the linguistic terms built with the '#' operator
%        that were found in source code files.
%      * A list with all errors and warnings generated by the parser.
%
%     OntologyFile can be '' (an empty string) if no ontology is
%     needed. The five lists returned by this predicate follow an
%     intermediate format intended to be read by 'translator' module
%     in order to build a TPL file.
%
%     @see translate_program/3
%

parse_program(ProgramFile, OntologyFile, Directives, Rules, Equations,
              LingTerms, Messages) :-
              
  % Sets the prefix for the user-defined predicates
  utilities:simplify_filename(ProgramFile, SimplifiedFilename),
  program_prefix(OldProgramPrefix),
  retract(program_prefix(OldProgramPrefix)),
  assert(program_prefix(SimplifiedFilename)),
  % Reads and parses program file
  reset_parser,
  parse_file(ProgramFile, DirectivesF, RulesF, EquationsF),
  % Checks if an ontology needs to be loaded
  (OntologyFile == '' ->
    DirectivesO = [], RulesO = [], EquationsO = [],
    (RulesF == [], 
     not((DirectivesF == [], EquationsF == [])),
     flags:get_bpl_flag(ontology_warning(true))
     ->
      add_message_in_file(ProgramFile,
        'Program have no rules and seems to be an ontology. \c
        Use \'ld -o\' if you want to load an ontology into a program.',
        warning)
    ;
      true
    )
  ;
    % Reads and parses ontology file
    parse_file(OntologyFile, DirectivesO, RulesO, EquationsO),
    (RulesO == [] ->
      true
    ;
      add_message_in_file(OntologyFile,
        'Ontologies can only contain equations and directives.',
          error)
    )
  ),
  % Sorts all errors and warnings by file name, line and column
  messages(UnsortedMessages),
  sort(UnsortedMessages, Messages),
  % Joins directives, equations and rules of both files
  append(DirectivesF, DirectivesO, Directives),
  append(RulesF, RulesO, Rules),
  append(EquationsF, EquationsO, Equations),
  % Gets the linguistic terms found in source code files and removes duplicates
  linguistic_terms(LingTermsWithDuplicates),
  list_to_set(LingTermsWithDuplicates, LingTerms).


%% parse_file(+File, -Directives, -Rules, -Equations)
%
%     Reads a Bousi-Prolog source file, performs a lexical, syntactic
%     and semantic analysis, and returns three lists with all the
%     directives, rules (including facts) and equations found. These
%     lists follow an intermediate format intended to be read by
%     'translator' module in order to build a TPL file.
%

parse_file(File, Directives, Rules, Equations) :-
  parse_file_aux(File, Directives, [], ReplacedDirectives, [], Rules, [], TmpEquations, []),
  % With the rules already read, generate proximity equations for the directive wn_gen_prox_equations in auto mode
  add_auto_gen_prox_equations(ReplacedDirectives, Rules, TmpEquations, Equations).
  
  
parse_file_aux(File, Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) :-
  % Saves source code filename, which will be used to show errors
  bpl_filename(OldFile),
  retract(bpl_filename(OldFile)),
  assert(bpl_filename(File)),
  % Opens file and reads its contents
  read_file_to_codes(File, Codes, []),
  atom_codes(Atom, Codes),
  % Parses source code
  foreign:ext_tokenize(Atom, Tokens),
  phrase(bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo), Tokens),
  !,
  writeln('Program parsed!'),
  % Restores original filename
  retract(bpl_filename(File)),
  assert(bpl_filename(OldFile)).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parser predicates for Bousi-Prolog queries
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% parse_query(+ProgramPrefix, +String, -Query, -NewHypoRules, -LingTerms, -Messages)
%
%     Reads the query contained in String atom, performs a lexical,
%     syntactic and semantic analysis, and (if no errors are found)
%     returns the following:
%
%      * A term that can be used to execute the query by means of the
%        'evaluator' module.
%      * A list with the linguistic terms built with the '#' operator
%        that were found in the query.
%      * A list with all errors and warnings generated by the parser.
%
%     Both the Query and the LingTerms list returned by this predicate
%     follow an intermediate format intended to be read by 'translator'
%     module.
%
%     If the query contains assumptions, they are returned in 
%     NewHypoRules.
%
%     @see translate_query/4
%

parse_query(ProgramPrefix, String, Query, NewHypoRules, LingTerms, Messages) :-
  % Sets the prefix for user-defined predicates
  program_prefix(OldProgramPrefix),
  retract(program_prefix(OldProgramPrefix)),
  assert(program_prefix(ProgramPrefix)),
  % Parses query
  reset_parser,
  foreign:ext_tokenize(String, Tokens),
  phrase(query(Query, NewHypoRules), Tokens),
  !,
  % Sorts all errors and warnings by file name, line and column
  messages(UnsortedMessages),
  sort(UnsortedMessages, Messages),
  % Gets the linguistic terms found in the query and removes duplicates
  linguistic_terms(LingTermsWithDuplicates),
  list_to_set(LingTermsWithDuplicates, LingTerms).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DCG (Definite Clause Grammar) rules for parsing Bousi-Prolog programs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% bousi_prolog_program(-DirectivesDiffList -ReplacedDirectivesDiffList 
%                       -RulesDiffList -EquationsDiffList)
%
%     DCG rule which parses a full Bousi-Prolog program and saves all
%     directives, rules (including facts) and fuzzy relation equations
%     in Directives, Rules and Equations lists, respectively. This rule
%     manages syntax errors and tries to continue parsing if one of
%     them is found, so it won't fail even if the Bousi-Prolog program
%     has errors.
%

% bousi_prolog_program(Directives, Rules, Equations) -->
%   bousi_prolog_program(Directives, [], Rules, [], Equations, []).


bousi_prolog_program(Dir, Dir, RDir, RDir, Rul, Rul, Equ, Equ) -->
  % Normal end-of-file
  [eof(_, _)].

bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Bousi-Prolog equation
  bpl_equation(Equi, Equ1, _),
  {
    !
  },
  bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equ1, Equo).
  
bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Bousi-Prolog specific directive
  % The directive wn_gen_prox_equations generates new proximity equations as well
  bpl_directive(Diri, Dir1, RDiri, RDir1, Equi, Equ1, _),
  {
    !
  },
  bousi_prolog_program(Dir1, Diro, RDir1, RDiro, Ruli, Rulo, Equ1, Equo).

bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Special include/1 directive
  include_directive(Diri, Dir1, RDiri, RDir1, Ruli, Rul1, Equi, Equ1, _),
  {
    !
  },
  bousi_prolog_program(Dir1, Diro, RDir1, RDiro, Rul1, Rulo, Equ1, Equo).

bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Prolog directive
  prolog_directive([Directive], _),
  {
    !,
    Diri = [Directive|Dir1]
  },
  bousi_prolog_program(Dir1, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo).

bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Rule or fact
  rule(Rules, [Line, _]),
  {
    !,
    % Rule weight is the degree of the graded rule: Rule with RuleWeight.
    % If the weight is below the lambda-cut, it was not loaded
    (Rules == []
     ->
      Rul1 = Ruli
    ;
      Rules = [[RuleContent, _RuleWeight, _HeadConstraints, _BodyConstraints, _HeadDegreeVars, _BodyDegreeVars, _HypoData]|_HypoRules],
      check_singleton(RuleContent, Line),
      append(Rules, Rul1, Ruli)
    )
  },
  bousi_prolog_program(Diri, Diro, RDiri, RDiro, Rul1, Rulo, Equi, Equo).

bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % If all of the previous rules fail, there must be a syntax error
  % in source code; this rule will try to find it and go on parsing
  bousi_prolog_program_error(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo).

  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% add_auto_gen_prox_equations(+Directives, +Rules, -InEquations, -OutEquations)
%%%
%%% If Directives is [:- directive(wn_gen_prox_equations, [Measure, Auto])], then 
%%% return in OutEquations all the equations derived from constants in Rules
%%% Otherwise, return InEquations

add_auto_gen_prox_equations([:- directive(Directive, [Measure, Auto])], Rules, InEquations, OutEquations) :-
  Directive==wn_gen_prox_equations,
  \+ is_list(Auto),
  !,
  wn_gen_prox_equations:wn_auto_gen_prox_equations([:- directive(Directive, [Measure, Auto])], Rules, InEquations, OutEquations).
  
add_auto_gen_prox_equations(_Directives, _Rules, Equations, Equations).


%% bpl_directive(-DirectivesIn, -DirectivesOut, -ReplacedDirectivesIn, -ReplacedDirectivesOut, -EquationsIn, -EquationsOut, -LineColumn)
%
%     DCG rule which parses a Bousi-Prolog specific directive and returns
%     it inside a directive/2 directive (e.g., ':- transitivity(yes).' is
%     translated into ':- directive(transitivity, [yes]).').
%
%     A directive, such as wn_gen_prox_equations, can be replaced by the 
%     proximity equations it generates.
%

bpl_directive(Diri, Diro, RDiri, RDiro, Equi, Equo, [Line, Column]) -->
  [name(':-', [Line, Column])], 
  term(Term, [], _, _), 
  [name('.', _)],
  {
    parse_expression(Term, ResultTerm),
    ResultTerm =.. [Name|Arguments],
    catch((
      % Checks if specified directive is a valid BPL directive
      directives:is_directive_valid(Name, Arguments),
      % Executes the directive
      directives:exec_directive(Name, Arguments, Equi, Equo),
      (directives:replaced_directive(Name) ->
         Diri = Diro,
         RDiri = [:- directive(Name, Arguments)|RDiro]
        ;
         Diri = [:- directive(Name, Arguments)|Diro],
         RDiri = RDiro)
    % (catcher)
    ), directive_error(ErrorMessage), (
      % Directive isn't valid
      add_message(ErrorMessage, Line, Column, error),
      Diri = Diro
    ))
  }.


%% bpl_equation(-EquationIn, -EquationOut, -LineColumn)
%
%     DCG rule which parses a Bousi-Prolog proximity, similarity or fuzzy
%     relation equation "a <op> b = n." (where <op> can be any of the
%     Bousi-Prolog relation operators) and returns it in Equation list as
%     a term with this syntax: "<name>(a, b, n)" (where <name> is the
%     internal name of <op>). <op> and <name> are related using
%     the relation_name/2 predicate of the 'utilities' module.
%

bpl_equation(Equi, Equo, [Line, Column]) -->
  symbol(Sym1, [Line, Column]), [relation(RelOp, _)], symbol(Sym2, _),
  [name('=', _)], number_(Value, [LineValue, ColumnValue]), [name('.', _)],
  {
    ((Value >= 0.0, Value =< 1.0)
     ->
      utilities:relation_name(RelOp, RelName),
      (RelName == sim,
       flags:get_bpl_flag(lambda_cut(LambdaCut)),
       LambdaCut > Value,
       flags:get_bpl_flag(filtering(Filtering)),
       Filtering==true
       ->
        (flags:get_bpl_flag(filtering_warning(true)) ->
           add_message('Equation not loaded: value is below the lambda cut.',
              LineValue, ColumnValue, warning)
          ;
           true
        ),
        Equi = Equo
      ;
       EqAux =.. [RelName, Sym1, Sym2, Value], 
       Equi = [EqAux|Equo]
       )
     ;
      add_message('Equation value is not a number in the range [0.0, 1.0].',
        LineValue, ColumnValue, error),
      Equi = Equo
    )
  }.


%% include_directive(-DirectivesIn, -DirectivesOut, 
%%                   -ReplacedDirectivesIn, -ReplacedDirectivesOut, 
%%                   -EquationsIn, -EquationsOut,
%%                   -RulesIn, -RulesOut, -LineColumn)
%
%     DCG rule which parses the special include/1 directive, reads the
%     included Bousi-Prolog source code file and return all directives,
%     rules and equations found in it.
%

include_directive(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo, [Line, Column]) -->
  [name(':-', [Line, Column])], atom_('include', _),
  [left_parenthesis(_, _)], atom_(QuotedFile, _), [right_parenthesis(_, _)],
  [name('.', _)],
  {
    % Gets the real path of the included file
    utilities:remove_quotes(QuotedFile, File),
    (is_absolute_file_name(File) ->
      TmpRealFile = File
    ;
      % Joins the path of the BPL file that is being parsed and the
      % included filename to get its real path; this is needed if the
      % BPL file isn't in the current working directory, for example
      % if '/home/user/pl/a.bpl' contains ':- include('b.bpl').' and
      % the file is loaded from '/home/user' using 'ld pl/a.bpl', the
      % real path of the included file isn't 'b.bpl', but 'pl/b.bpl'
      bpl_filename(CurrentFile),
      file_directory_name(CurrentFile, CurrentDirectory),
      concat_atom([CurrentDirectory, '/', File], TmpRealFile)
    ),
    (exists_file(TmpRealFile) ->
      RealFile = TmpRealFile
     ;
      atom_concat(TmpRealFile,'.bpl',RealFile)
    ),
    % Checks whether the included file exists
    (exists_file(RealFile) ->
      % Parses recursively the included file
      parse_file_aux(RealFile, Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo)
    ;
      % Included file doesn't exist or path is invalid
      swritef(Message, 'Included file does not exist: \'%w\'.', [TmpRealFile]),
      add_message(Message, Line, Column, error),
      Diri = Diro,
      RDiri = RDiro,
      Ruli = Rulo, 
      Equi = Equo
    )
  }.


%% prolog_directive(-Directive, -LineColumn)
%
%     DCG rule which parses a Prolog directive and returns it in
%     Directive list. If the parsed directive declares new operators,
%     this predicate will add them to custom operator list to allow
%     them to be used in following statements.
%

prolog_directive([Directive], [Line, Column]) -->
  [name(':-', [Line, Column])], 
  term(Term, [], _, _), 
  [name('.', _)],
  {
    parse_expression(Term, ResultTerm),
    % Checks if this directive needs a special treatment
    (ResultTerm = op(OpPrio, OpType, OpName) ->
      (atom(OpName) ->
        % Directive declares a single operator
        declare_custom_operator(OpPrio, OpType, [OpName])
      ;
        % Directive declares a list of operators
        declare_custom_operator(OpPrio, OpType, OpName)
      ),
      Directive = (:- ResultTerm)
    ;
    (ResultTerm = dynamic(Pred/Arity) ->
      % Directive declares a dynamic predicate, so its name and
      % arity must be changed from 'pred/2' to 'prefix_pred/3'
      program_prefix(Prefix),
      concat_atom([Prefix, '_', Pred], NewPred),
      translator:expanded_rule_arity(Arity, NewArity),
      Directive = (:- dynamic(NewPred/NewArity))
    ;
    (ResultTerm = discontiguous(Pred/Arity) ->
      % Directive declares a dynamic predicate, so its name and
      % arity must be changed from 'pred/2' to 'prefix_pred/3'
      program_prefix(Prefix),
      concat_atom([Prefix, '_', Pred], NewPred),
      translator:expanded_rule_arity(Arity, NewArity),
      Directive = (:- discontiguous(NewPred/NewArity))
    ;
      % No special treatment needed
      Directive = (:- ResultTerm)
    )))
  }.


%% rule(-Rule, -LineColumn)
%
%     DCG rule which parses a Prolog/Bousi-Prolog clause and converts
%     it into a list with an intermediate format that can be later
%     used by the 'translator' module. The returned list will have the
%     following items:
%
%      * A copy of the original clause in which the rule's head and
%        each call to a user-defined predicate has been modified in
%        order to add a degree variable and the prefix of the program
%        being compiled.
%      * Lists of head and body constraint variables added to 
%        the rule's body.
%      * Lists of head and body degree variables added to 
%        the rule's body.
%      * Hypothetical data.
%
%     For example, if the program being loaded is called "prog.bpl",
%     given the rule "a(X) :- b(X), c, true", the parser will translate
%     this rule into a list with these items:
%
%      * a) prog_a(X, Ci, Co, DG) :-
%             prog_b(X, Ci, C1, D1), 
%             prog_c(C1, Co, D2), 
%             true.
%      * b) [DG]
%      * c) [D1, D2]
%      * d) [Ci, Co]
%      * e) [C1]
%      * f) hypo(RuleVars, CD)
%
%   New hypothetical rules can come from:
%      1- Adding hypo data when transforming the rule.
%      2- Parsing higher-order predicates with term//4.
%

%  term//3 creates a list of elements in the term
%  parse_expression/2 converts this list into a Prolog term
%  translate_rule adds, by means of the scanners:
%     Degree variable(s): one or two variables depending on degree 
%       accumulator optimization.
%     Block constraint variables: two variables if either a2 or a3
%       weak unification algorithms are selected.
%     The hypothetical term if hypothetical reasoning is enabled. 
%       In this case, several rules may be returned for each single
%       translated rule.

rule(Rules, [Line, Column]) -->
  term(RuleTerm, [], NewHypoRules, [Line, Column]),
  [name('.', _)], % End of a rule
  { parse_and_translate_rule(RuleTerm, [Line, Column], RuleRules),
    append(RuleRules, NewHypoRules, Rules) }.


%% parse_and_translate_rule(+RuleTerm, +Location, -Rules)
%     Translates the list of terms and operators returned by term/2
%     into a real Prolog term
%
parse_and_translate_rule(RuleTerm, Location, Rules) :-
  parse_expression(RuleTerm, RulePrologTerm),
  translate_rule(RulePrologTerm, Location, Rules).


%% translate_rule(+RulePrologTerms, +Location, -Rules)
%
%    Translates a parsed rule (a Prolog term) into a list of rules, i.e.,
%    adding the prefix to defined predicates, and adding variables
%    for the approximation degree, constraints and the hypothetical
%    term.
%
%    Rule expansion is delayed up to the translator module
%

translate_rule(RulePrologTerm, Location, Rules) :-
    (RulePrologTerm = with(_Rule, _RuleWeight)
     -> GradedRule = RulePrologTerm
     ;  GradedRule = with(RulePrologTerm, 1.0)),
     init_hypothetical_term(HypoTerm),
     % Translate the parsed rule (variables are still atoms and are
     % converted here to Prolog variables)
    translate_rule(GradedRule, HypoTerm, Location, Rules).


%% translate_rule(+Rule, +RuleWeight, +HypoTerm, +Location, -Rules)
%
%    Translate a parsed rule by adding:
%      - The hidden accumulator (if it is a DCG)
%      - Degree variable(s) (one or two, if degree accumulator 
%        optimization is enabled)
%      - Adding constraint variables (for a2 and a3)
%      - Adding the hypothetical term (if hypothetical reasoning 
%        is enabled)
%
%    The rule is rejected if it is below the lambda-cut
%
%    HypoTerm = hypo(SharedVs, RI, RD, CD)

translate_rule(Rule, HypoTerm, Location, Rules) :-
  translate_rule(Rule, _RuleWeight, HypoTerm, Location, Rules).

translate_rule(Rule, RuleWeight, HypoTerm, [Line, Column], Rules) :-
  (flags:get_bpl_flag(hypothetical(true))
   -> % Linearize SharedVs by creating a fresh list of variables
      % which will be weakly unified in the rule body pairwise
      HypoTerm = hypo(SharedVs, RI, RD, CD),
      length(SharedVs, NbrSharedVs),
      length(NewSharedVs, NbrSharedVs),
      get_ri(RI)
   ;  SharedVs = [],
      NewSharedVs = []),
  rule_weight(Rule, RuleWOWeight, RuleWeight),
  \+ atom_is_variable(RuleWOWeight),
  (
    % A DCG
    RuleWOWeight = (HeadTerm --> BodyWithoutDegreeVars),
    !,
    valid_parsed_head(HeadTerm),
    check_builtin_redefinition(HeadTerm, Line, Column),
    dcg_translation('-->'(HeadTerm, BodyWithoutDegreeVars), ':-'(HeadWithoutDegreeVar, TrBodyWithoutDegreeVars)),
    parsed_term_variables([HeadTerm, BodyWithoutDegreeVars], RuleVars),
    translate_rule_body(TrBodyWithoutDegreeVars, Body, BodyDegreeVars, BodyConstraintBlockVars, hypo(body, RuleVars, CD), NewHypoRules)
  ;
    % A rule
    RuleWOWeight = (HeadTerm :- BodyWithoutDegreeVars),
    !,
    valid_parsed_head(HeadTerm),
    check_builtin_redefinition(HeadTerm, Line, Column),
    HeadWithoutDegreeVar = HeadTerm,
    parsed_term_variables([HeadTerm, BodyWithoutDegreeVars], RuleVars),
    translate_rule_body(BodyWithoutDegreeVars, Body, BodyDegreeVars, BodyConstraintBlockVars, hypo(body, RuleVars, CD), NewHypoRules)
  ;
    % A fact
    % Rule = (HeadTerm :- true),
    RuleWOWeight = HeadTerm,
    valid_parsed_head(HeadTerm),
    HeadWithoutDegreeVar = HeadTerm,
    check_builtin_redefinition(HeadTerm, Line, Column),
    Body = true,
    BodyDegreeVars = [],
    BodyConstraintBlockVars = [],
    NewHypoRules = []
  ),
  !,
  % Checks that rule weight is over lambda-cut
  flags:get_bpl_flag(lambda_cut(LambdaCut)),
  flags:get_bpl_flag(filtering(Filtering)),
  (Filtering == true,
   RuleWeight < LambdaCut,
    % Do not add the rule
    (
     flags:get_bpl_flag(filtering_warning(true)) ->
       add_message('Rule not loaded: rule weight is below the lambda cut.',
          Line, Column, warning)
      ;
       true),
     Rules = []
  ;
    % Add the rule
    % Adds a degree variable to rule's head
    add_degree_variables(HeadWithoutDegreeVar, HeadWithDegreeVar, HeadDegreeVars),
    % Adds the constraint variables to rule's head
    add_constraint_variables(HeadWithDegreeVar, HeadWithConstraintsAndDegreeVar, HeadConstraintBlockVars),
    % Adds the initial hypothetical term to rule's head
    add_hypothetical_term(HeadWithConstraintsAndDegreeVar, Head, hypo(head, HypoTerm), _EmptyList),
    % Builds a list with the parsed rule, rule weight, head constraint
    % body constraints, head degree, body degree, and hypothetical data
    % created while scanning its body
    Rules = [[(Head :- Body), RuleWeight, HeadConstraintBlockVars, BodyConstraintBlockVars, HeadDegreeVars, BodyDegreeVars, hypo(SharedVs, NewSharedVs, RI, RD, CD)]|NewHypoRules]
  ).

%% translate_rule_body(+BodyWithoutDegreeVars, -Body, -BodyDegreeVars, -BodyConstraintBlockVars, +HypoData, -NewHypoRules)
%
%    Adds degree variables, constraints, and hypo term to the body
%     HypoData = hypo(body, RuleVars, CD)
%

translate_rule_body(Body, NewBody, BodyDegreeVars, BodyConstraintBlockVars, HypoData, NewHypoRules) :-
  % Replace variable calls
  replace_variable_calls(Body, BodyWithoutDegreeVars), 
  % Registers whter Body contains a disjunction. If not, apply a simplified translation
  %   in add_degree_variables/3 and add_constraint_variables/3
  register_disjunction_in_term(Body),
  % Adds the degree variables to rule's body
  add_degree_variables(BodyWithoutDegreeVars, BodyWithDegreeVars, BodyDegreeVars),
  % Adds the constraint variables to rule's body
  add_constraint_variables(BodyWithDegreeVars, BodyWithCtrsAndDegreeVars, BodyConstraintBlockVars),
  % Adds the hypothetical term to rule's body
  add_hypothetical_term(BodyWithCtrsAndDegreeVars, NewBody, HypoData, NewHypoRules).

%% check_builtin_redefinition(+HeadWithoutDegreeVar, +Line, +Column) 
%
%     Check whether a given predicate is already a built-in. If so, raise an error
%
check_builtin_redefinition(HeadWithoutDegreeVar, Line, Column) :-
  % Checks if the rule's head references an existing predicate
  (utilities:builtin(HeadWithoutDegreeVar) ->
    functor(HeadWithoutDegreeVar, Functor, Arity),
    swritef(Message, 'No permission to redefine built-in predicate \'%w/%w\'.', [Functor, Arity]),
    add_message(Message, Line, Column, error)
  ;
    true
  ).
  
   
%% rule_weight(+Rule, -RuleWOWeight, -RuleWeight)
%
%     Extracts the rule weight of a rule

rule_weight(with(Rule, RuleWeight), Rule, RuleWeight) :-
  !.

% rule_weight(with((Head --> Body), RuleWeight), (Head --> Body), RuleWeight) :-
%   !.

% rule_weight(with(Head, RuleWeight), Head, RuleWeight) :-
%   !.

rule_weight(Rule, Rule, 1.0) :-
  !.


%% term(-Term, +Options, -NewHypoRules, -LineColumn)
%
%     DCG rule which parses a Prolog/Bousi-Prolog atomic or compound
%     term and returns each of its components (operands and operators)
%     in a list. Operators are stored in terms with this syntax:
%     'exp_operator(<operator>, <prefix/infix>, [<line>, <column>])'.
%
%     Options is a list that can contain the following items:
%
%      * not_allowed_ops(+Operators): Operators is a list of operators
%        that can't be used as the main operator in this term.
%      * no_comparisons: indicates that the parsed term can't be a
%        Bousi-Prolog weak unification or term comparison.
%

%% remaining_term(-RemainingTerm, +Options, -NewHypoRules)
%
%     Helper DCG rule used by term//4 to reduce backtracking and 
%     increase analysis speed.
%
%     @see term//4
%

term(Term, Options, NewHypoRules, [Line, Column]) -->
  % Term with a prefix operator (operator can't be '\+' nor 'not'
  % because negations are handled in basic_prolog_term//3 rule)
  operator(Operator, [Line, Column]),
  {
    % Checks that operator is allowed
    Operator \== \+, Operator \== not,
    (member(not_allowed_ops(NotAllowedOps), Options) ->
      not(member(Operator, NotAllowedOps))
    ;
      true
    ),
    % Verifies that operator is defined and it's a prefix one
    operator_type(_, OpType, Operator),
    member(OpType, [fx, fy])
  },
  term(SingleTerm, Options, NewHypoRules, _),
  {
    Term = [exp_operator(Operator, prefix, [Line, Column])|SingleTerm]
  }.

term(Term, Options, NewHypoRules, [Line, Column]) -->
  % Term with an infix operator, or without operator
  basic_term(FirstTerm, Options, NewHypoRules1, [Line, Column]),
  remaining_term(RemainingTerm, Options, NewHypoRules2),
  {
    append([FirstTerm], RemainingTerm, Term),
    append(NewHypoRules1, NewHypoRules2, NewHypoRules)
  }.

remaining_term(RemainingTerm, Options, NewHypoRules) -->
  % Infix operator
  operator(Operator, [LineOp, ColumnOp]),
  {
    (member(not_allowed_ops(NotAllowedOps), Options) -> 
      not(member(Operator, NotAllowedOps))
    ;
      true
    ),
    operator_type(_, OpType, Operator),
    member(OpType, [xfx, xfy, yfx])
  },
  term(Term, Options, NewHypoRules, _),
  {
    RemainingTerm = [exp_operator(Operator, infix, [LineOp, ColumnOp])|Term]
  }.

remaining_term([], _Options, []) -->
  % No operator
  [].


%% basic_term(-Term, +Options, -NewHypoRules, -LineColumn)
%
%     DCG rule which parses and returns a Prolog/Bousi-Prolog term
%     without operators. For a list of valid Options, see term/3.
%
%     @see term/3
%

basic_term(RelName, _Options, [], [Line, Column]) -->
  % Single BPL operator
  [relation(RelOperator, [Line, Column])],
  {
    utilities:relation_name(RelOperator, RelName)
  }.

basic_term(Term, _Options, NewHypoRules, [Line, Column]) -->
  % Prolog term
  basic_prolog_term(Term, NewHypoRules, [Line, Column]).

basic_term(BPLTerm, Options, [], [Line, Column]) -->
  % Bousi-Prolog weak unification or term comparison
  {
    not(member(no_comparisons, Options))
  },
  bpl_comparison(BPLTerm, [Line, Column]).


%% bpl_comparison(-BPLTerm, -LineColumn)
%
%     DCG rule which parses and returns a Bousi-Prolog weak unification
%     (e.g. "a(X) ~ b(X) < 0.5") or term comparison (e.g. "a ~> b =:= 1").
%

bpl_comparison(BPLTerm, [Line, Column]) -->
  % Weak unification operator (~)
  basic_term(BasicTerm1, [no_comparisons], _, [Line, Column]),
  [relation('~', _)],
  basic_term(BasicTerm2, [no_comparisons], _, _),
  [name(Operator, _)], number_or_variable(Value, _),
  {
    member(Operator, [=, =:=, =\=, =<, >=, >, <]),
    % A term as "a(X) ~ b(X) >= 0.5" is translated into:
    % "unify_a1(a(X), b(X), >=, 0.5)"
    unify_predicate_name(UnifyPredicateName) ,
    BPLTerm =.. [UnifyPredicateName, BasicTerm1, BasicTerm2, Operator, Value]
  }.

bpl_comparison(BPLTerm, [Line, Column]) -->
  % Syntactic sugar: "a ~ b" is equivalent to "a ~ b > 0"
  basic_term(BasicTerm1, [no_comparisons], _, [Line, Column]),
  [relation('~', _)],
  basic_term(BasicTerm2, [no_comparisons], _, _),
  {
    unify_predicate_name(UnifyPredicateName),
    BPLTerm =.. [UnifyPredicateName, BasicTerm1, BasicTerm2, >, 0]
  }.

bpl_comparison(BPLTerm, [Line, Column]) -->
  % Weak unification operator (~~)
  basic_term(BasicTerm1, [no_comparisons], _, [Line, Column]),
  [relation('~~', _)],
  basic_term(BasicTerm2, [no_comparisons], _, _),
  {
    % A term as "a(X) ~~ b(X)" is translated into:
    % "weak_unify_a1(a(X), b(X), D)" for algorithm A1
    % "weak_unify_a3(a(X), b(X), Lambda, Cin, Cout, D)" for algorithm A3
    weak_unify_predicate_name(UnifyPredicateName),
    flags:get_bpl_flag(lambda_cut(LambdaCut)),
    BPLTerm =.. [UnifyPredicateName, BasicTerm1, BasicTerm2, LambdaCut]
    % Cin, Cout, and Degree will be added later (in rule/4)
  }.

bpl_comparison(BPLTerm, [Line, Column]) -->
  % Extended BPL operators (~>, <~, ~1~, ~2~, ~3~)
  basic_term(BasicTerm1, [no_comparisons], _, [Line, Column]),
  [relation(RelOperator, _)],
  basic_term(BasicTerm2, [no_comparisons], _, _),
  [name(Operator, _)], number_or_variable(Value, _),
  {
    RelOperator \== '~',
    utilities:relation_name(RelOperator, RelName),
    utilities:relation_evaluator(RelName, RelEvaluator),
    member(Operator, [=, =:=, =\=, =<, >=, >, <]),
    % A term like "a(X) ~2~ b(X) =:= 1.0" is translated into:
    % "e_frel2(a(X), b(X), =:=, 1.0)"
    BPLTerm =.. [RelEvaluator, BasicTerm1, BasicTerm2, Operator, Value]
  }.

bpl_comparison(BPLTerm, [Line, Column]) -->
  % Syntactic sugar: "a · b" is equivalent to "a · b > 0"
  basic_term(BasicTerm1, [no_comparisons], _, [Line, Column]),
  [relation(RelOperator, _)],
  basic_term(BasicTerm2, [no_comparisons], _, _),
  {
    RelOperator \== '~',
    utilities:relation_name(RelOperator, RelName),
    utilities:relation_evaluator(RelName, RelEvaluator),
    BPLTerm =.. [RelEvaluator, BasicTerm1, BasicTerm2, >, 0]
  }.


unify_predicate_name(UnifyPredicateName) :-
    flags:get_bpl_flag(weak_unification(Algorithm)),
    atom_concat('unify_', Algorithm, UnifyPredicateName).
    
weak_unify_predicate_name(WeakUnifyPredicateName) :-
    unify_predicate_name(UnifyPredicateName),
    atom_concat('weak_', UnifyPredicateName, WeakUnifyPredicateName).


%% basic_prolog_term(-Term, -NewHypoRules, -LineColumn)
%
%     DCG rule which parses and returns a Prolog atomic term or a
%     Prolog compound term with functional notation. This rule will
%     also consider Bousi-Prolog linguistic terms built with the '#'
%     operator as valid atoms.
%

basic_prolog_term(Number, [], [Line, Column]) -->
  % Integer number
  [integer(Number, [Line, Column])].

basic_prolog_term(Number, [], [Line, Column]) -->
  % Floating point number
  [float(Number, [Line, Column])].

basic_prolog_term(Variable, [], [Line, Column]) -->
  % Variable
  [variable(Variable, [Line, Column])].

basic_prolog_term(HigherOrderTerm, NewHypoRules, [Line, Column]) -->
  % Higher-order predicate (includes negated terms)
  higher_order_prolog_term(HigherOrderTerm, NewHypoRules, [Line, Column]).

basic_prolog_term(Compound, [], [Line, Column]) -->
  % Compound term (functional notation)
  atom_or_linguistic_term(Atom, [Line, Column]),
  [left_parenthesis(_, _)], argument_list(Args, _), [right_parenthesis(_, _)],
  {
    Compound =.. [Atom|Args]
  }.

basic_prolog_term(InnerTerm, NewHypoRules, [Line, Column]) -->
  % Term in parenthesis
  [left_parenthesis(_, [Line, Column])], 
  term(Term, [], NewHypoRules, _), 
  [right_parenthesis(_, _)],
  {
    parse_expression(Term, InnerTerm)
  }.

basic_prolog_term({InnerTerm}, NewHypoRules, [Line, Column]) -->
  % Term in curly braces
  [left_curly_brace(_, [Line, Column])], 
  term(Term, [], NewHypoRules, _), 
  [right_curly_brace(_, _)],
  {
    parse_expression(Term, InnerTerm)
  }.

basic_prolog_term(List, [], [Line, Column]) -->
  % Non-empty list
  [left_bracket(_, [Line, Column])], items(List, _), [right_bracket(_, _)].

basic_prolog_term(CharCodeList, [], [Line, Column]) -->
  % String as a character code list
  [character_code_list(CharCodeList, [Line, Column])].

basic_prolog_term(Atom, [], [Line, Column]) -->
  % Atom (including empty list)
  atom_or_linguistic_term(Atom, [Line, Column]).


%% bpl_linguistic_term(-BPLTerm, -LineColumn)
%
%     DCG rule which parses a Bousi-Prolog linguistic term built with
%     the '#' operator, creates the definition of the associated fuzzy
%     subset (using the notation required by ext_translate_fuzzysets/5),
%     and returns the simplified name of the subset. A reference to the
%     subset is also saved in the linguistic_terms/1 dynamic predicate.
%

bpl_linguistic_term(BPLTerm, [Line, Column]) -->
  % 1st type: crisp domain ranges
  % domain#minimum#maximum (e.g., height#20#100)
  [name(Domain, [Line, Column])], [builder(_, _)],
  integer_number(Minimum, _), [builder(_, _)],
  integer_number(Maximum, _),
  {
    % Builds the name and the definition of the fuzzy subset
    % 'height#20#100' is translated into 'height_20_100(between(20, 100))'
    concat_atom([Domain, '_', Minimum, '_', Maximum], InitialBPLTerm),
    utilities:simplify_atom(InitialBPLTerm, BPLTerm),
    NewSubset =.. [BPLTerm, between(Minimum, Maximum)],
    % Adds the new subset to the list of linguistic terms
    linguistic_terms(OldLingTerms),
    retract(linguistic_terms(OldLingTerms)),
    append(OldLingTerms, [[domain, Domain, NewSubset]], NewLingTerms),
    assert(linguistic_terms(NewLingTerms))
  }.

bpl_linguistic_term(BPLTerm, [Line, Column]) -->
  % 2nd type: fuzzy domain ranges
  % modifier#domain#minimum#maximum (e.g., about#speed#100#120)
  % Valid modifiers: [about]
  [name(Modifier, [Line, Column])], [builder(_, _)],
  [name(Domain, _)], [builder(_, _)],
  integer_number(Minimum, _), [builder(_, _)],
  integer_number(Maximum, _),
  {
    % Builds the name and the definition of the fuzzy subset
    % 'about#speed#100#120' is translated into
    % 'about_speed_100_120(about(100, 120))'
    member(Modifier, [about]),
    concat_atom([Modifier, '_', Domain, '_', Minimum, '_', Maximum], InitialBPLTerm),
    utilities:simplify_atom(InitialBPLTerm, BPLTerm),
    NewSubsetDef =.. [Modifier, Minimum, Maximum],
    NewSubset =.. [BPLTerm, NewSubsetDef],
    % Adds the new subset to the list of linguistic terms
    linguistic_terms(OldLingTerms),
    retract(linguistic_terms(OldLingTerms)),
    append(OldLingTerms, [[domain, Domain, NewSubset]], NewLingTerms),
    assert(linguistic_terms(NewLingTerms))
  }.

bpl_linguistic_term(BPLTerm, [Line, Column]) -->
  % 3rd type: domain points
  % domain#value (e.g., temperature#38)
  [name(Domain, [Line, Column])], [builder(_, _)],
  integer_number(Number, _),
  {
    % Builds the name and the definition of the fuzzy subset
    % 'temperature#38' is translated into 'temperature_38(point(38))'
    concat_atom([Domain, '_', Number], InitialBPLTerm),
    utilities:simplify_atom(InitialBPLTerm, BPLTerm),
    NewSubset =.. [BPLTerm, point(Number)],
    % Adds the new subset to the list of linguistic terms
    linguistic_terms(OldLingTerms),
    retract(linguistic_terms(OldLingTerms)),
    append(OldLingTerms, [[domain, Domain, NewSubset]], NewLingTerms),
    assert(linguistic_terms(NewLingTerms))
  }.

bpl_linguistic_term(BPLTerm, [Line, Column]) -->
  % 4th type: fuzzy domain points
  % modifier#domain#value (e.g., about#age#50)
  % Valid modifiers: [about]
  [name(Modifier, [Line, Column])], [builder(_, _)],
  [name(Domain, _)], [builder(_, _)],
  integer_number(Value, _),
  {
    % Builds the name and the definition of the fuzzy subset
    % 'about#age#50' is translated into 'about_age_50(about(50))'
    member(Modifier, [about]),
    concat_atom([Modifier, '_', Domain, '_', Value], InitialBPLTerm),
    utilities:simplify_atom(InitialBPLTerm, BPLTerm),
    NewSubsetDef =.. [Modifier, Value],
    NewSubset =.. [BPLTerm, NewSubsetDef],
    % Adds the new subset to the list of linguistic terms
    linguistic_terms(OldLingTerms),
    retract(linguistic_terms(OldLingTerms)),
    append(OldLingTerms, [[domain, Domain, NewSubset]], NewLingTerms),
    assert(linguistic_terms(NewLingTerms))
  }.


bpl_linguistic_term(BPLTerm, [Line, Column]) -->
  % 5th type: compound linguistic terms
  % modifier#subset (e.g., extremely#small)
  % Valid modifiers: [very, somewhat, more_or_less, extremely]
  [name(Modifier, [Line, Column])], [builder(_, _)],
  [name(Subset, _)],
  {
    % Builds the name and the definition of the fuzzy subset
    % 'extremely#small' is translated into 'extremely_small(extremely(small))'
    member(Modifier, [very, somewhat, more_or_less, extremely]),
    concat_atom([Modifier, '_', Subset], InitialBPLTerm),
    utilities:simplify_atom(InitialBPLTerm, BPLTerm),
    NewSubsetAux =.. [Modifier, Subset],
    NewSubset =.. [BPLTerm, NewSubsetAux],
    % Adds the new subset to the list of linguistic terms
    linguistic_terms(OldLingTerms),
    retract(linguistic_terms(OldLingTerms)),
    append(OldLingTerms, [[subset, Subset, NewSubset]], NewLingTerms),
    assert(linguistic_terms(NewLingTerms))
  }.


%% higher_order_prolog_term(-Compound, -NewHypoRules, -LineColumn)
%
%     DCG rule which parses a Prolog higher-order predicate. These
%     predicates requires a special treatment because Bousi-Prolog
%     adds a prefix to the names of the predicates of the program, so
%     some changes must be done to the arguments that define a goal.
%
%     Each argument of these predicates that is a goal can be a
%     partial goal, which means that it needs run-time compilation.
%
%     Currently, these are the supported higher-order predicates:
%
%      * not(:Goal)
%      * \+(:Goal)
%      * assert(:Goal)                   - Crisp
%      * retract(:Goal)                  - Crisp
%      * =>(Rules, :Goal)
%      * call(:Goal)
%      * once(:Goal)
%      * ignore(:Goal)
%      * time(:Goal)
%      * truth_degree(:Goal, -Number)    - Crisp
%      * apply(:Goal, +List)
%      * maplist(:Goal, +List)
%      * forall(:Condition, :Action)     - Crisp
%      * findall(+Template, :Goal, -Bag) - Crisp
%      * bagof(+Template, :Goal, -Bag)   - Crisp
%      * setof(+Template, :Goal, -Bag)   - Crisp
%      * catch(:Goal, +Catcher, :Recover)
%
%     Goals can be any kind of term, including compound terms built
%     with control predicates such as ,/2, ;/2 or ->/2. 
%

higher_order_prolog_term(Compound, [], [Line, Column]) -->
  % Negation
  atom_(Atom, [Line, Column]),
  {
    (Atom == not ; Atom == \+)
  },
  % Allow 'not' and \+ to be used as metapredicates and prefix operators
  ( [left_parenthesis(_, _)]
    -> term(Term, [not_allowed_ops([','])], _NewHypoRules1, _),
       [right_parenthesis(_, _)]
    ;  term(Term, [not_allowed_ops([','])], _NewHypoRules1, _)
  ),
  {
    parse_expression(Term, ResultTerm),
    Compound = eval_negation(Atom, ResultTerm)
  }.

higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order crisp predicates with signature "name(:Rule)":
  % assert/1, retract/1
  atom_(Atom, [Line, Column]),
  {
    higher_order_predicate(Atom, [atom])
  },
  [left_parenthesis(_, _)],
  term(RuleTerm, [], NewHypoRules, _),
  [right_parenthesis(_, _)],
  {
    % (RuleTerm = [VariableTerm],
    %  atom_is_variable(VariableTerm)
    (RuleTerm = [Rule],
     \+ is_total_rule(Rule)
     ->
      % A partial rule as argument
      Compiled = no,
      RuleRules = RuleTerm
     ;
      % A total rule as argument
      Compiled = yes,
      parse_and_translate_rule(RuleTerm, [Line, Column], RuleRules) ),
    Compound =.. [Atom, Compiled, RuleRules]
  }.

higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order predicates with signature "name(:Goal)":
  % call/1, once/1, ignore/1, time/1
  atom_(Atom, [Line, Column]),
  {
    higher_order_predicate(Atom, [goal]), Atom \== \+, Atom \== not,
    % 'call' should receive a new name because there are N predefined
    % call/N variants in SWI-Prolog 
    (Atom == call -> NewAtom = bpl_call ; NewAtom = Atom)
  },
  [left_parenthesis(_, _)],
  term(Term, [not_allowed_ops([','])], NewHypoRules, _),
  [right_parenthesis(_, _)],
  {
    parse_expression(Term, ResultTerm),
    Compound =.. [NewAtom, ResultTerm]
  }.

higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order predicates with signature "name(:Goal, +List)":
  % apply/2, maplist/2 
  atom_(Atom, [Line, Column]),
  {
    higher_order_predicate(Atom, [nongoal, nongoal])
  },
  [left_parenthesis(_, _)],
  term(Term1, [not_allowed_ops([','])], NewHypoRules1, _),
  [comma(_, _)],
  term(Term2, [not_allowed_ops([','])], NewHypoRules2, _),
  [right_parenthesis(_, _)],
  {
    % Parses the inner terms
    parse_expression(Term1, ResultTerm1),
    parse_expression(Term2, ResultTerm2),
    % Creates the final compound term
    (Atom == apply ->
      Compound =.. [bpl_apply, ResultTerm1, ResultTerm2]
    ; (Atom == maplist ->
      Compound =.. [bpl_maplist, ResultTerm1, ResultTerm2]
    ;
      Compound =.. [Atom, ResultTerm1, ResultTerm2]
    )),
    append(NewHypoRules1, NewHypoRules2, NewHypoRules)
  }.

higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order predicates with signature "name(:Goal, +Number)":
  % truth_degree/2 
  % Higher-order predicates with signature "name(:Condition, :Action)":
  % forall/2
  atom_(Atom, [Line, Column]),
  {
    higher_order_predicate(Atom, [goal, nongoal]) ;
    higher_order_predicate(Atom, [goal, goal])
  },
  [left_parenthesis(_, _)],
  term(Term1, [not_allowed_ops([','])], NewHypoRules1, _),
  [comma(_, _)],
  term(Term2, [not_allowed_ops([','])], NewHypoRules2, _),
  [right_parenthesis(_, _)],
  {
    % Parses the inner terms
    parse_expression(Term1, ResultTerm1),
    parse_expression(Term2, ResultTerm2),
    % Creates the final compound term
    Compound =.. [Atom, ResultTerm1, ResultTerm2],
    append(NewHypoRules1, NewHypoRules2, NewHypoRules)
  }.

higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order predicates with signature "name(+Template, :Goal, -Bag)":
  % findall/3, bagof/3, setof/3
  atom_(Atom, [Line, Column]),
  {
    higher_order_predicate(Atom, [nongoal, goal, nongoal])
  },
  [left_parenthesis(_, _)],
  term(Term1, [not_allowed_ops([','])], _, _),
  [comma(_, _)],
  term(Term2, [not_allowed_ops([','])], NewHypoRules, _),
  [comma(_, _)],
  term(Term3, [not_allowed_ops([','])], _, _),
  [right_parenthesis(_, _)],
  {
    % Parses the inner terms
    parse_expression(Term1, ResultTerm1),
    parse_expression(Term2, ResultTerm2),
    parse_expression(Term3, ResultTerm3),
    % Creates the final compound term
    Compound =.. [Atom, ResultTerm1, ResultTerm2, ResultTerm3]
  }.

higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order predicates with signature "name(+Goal, :List)":
  % phrase/2
  % built-in phrase/2 is replaced by my_phrase/3 (with its third arg. set to [])
  atom_(Atom, [Line, Column]),
  {
    Atom == phrase,
    higher_order_predicate(Atom, [goal, nongoal])
  },
  [left_parenthesis(_, _)],
  term(Term1, [not_allowed_ops([','])], NewHypoRules, _),
  [comma(_, _)],
  term(Term2, [not_allowed_ops([','])], _, _),
  [right_parenthesis(_, _)],
  {
    % Parses the inner terms
    parse_expression(Term1, ResultTerm1),
    parse_expression(Term2, ResultTerm2),
    % Creates the final compound term
    Compound =.. [my_phrase, ResultTerm1, ResultTerm2, []]
  }.
  
higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order predicates with signature "name(+Goal, :List, :List)":
  % phrase/3
  % built-in phrase/3 is replaced by my_phrase/3
  atom_(Atom, [Line, Column]),
  {
    Atom == phrase,
    higher_order_predicate(Atom, [goal, nongoal, nongoal])
  },
  [left_parenthesis(_, _)],
  term(Term1, [not_allowed_ops([','])], NewHypoRules, _),
  [comma(_, _)],
  term(Term2, [not_allowed_ops([','])], _, _),
  [comma(_, _)],
  term(Term3, [not_allowed_ops([','])], _, _),
  [right_parenthesis(_, _)],
  {
    % Parses the inner terms
    parse_expression(Term1, ResultTerm1),
    parse_expression(Term2, ResultTerm2),
    parse_expression(Term3, ResultTerm3),
    % Creates the final compound term
    Compound =.. [my_phrase, ResultTerm1, ResultTerm2, ResultTerm3]
  }.

higher_order_prolog_term(Compound, NewHypoRules, [Line, Column]) -->
  % Higher-order predicates with signature "name(:Goal, +Catcher, :Recover)":
  % catch/3
  atom_(Atom, [Line, Column]),
  {
    higher_order_predicate(Atom, [goal, nongoal, goal])
  },
  [left_parenthesis(_, _)],
  term(Term1, [not_allowed_ops([','])], NewHypoRules1, _),
  [comma(_, _)],
  term(Term2, [not_allowed_ops([','])], _, _),
  [comma(_, _)],
  term(Term3, [not_allowed_ops([','])], NewHypoRules2, _),
  [right_parenthesis(_, _)],
  {
    % Parses the inner terms
    parse_expression(Term1, ResultTerm1),
    parse_expression(Term2, ResultTerm2),
    parse_expression(Term3, ResultTerm3),
    % Creates the final compound term
    Compound =.. [Atom, ResultTerm1, ResultTerm2, ResultTerm3],
    append(NewHypoRules1, NewHypoRules2, NewHypoRules)
  }.


%% argument_list(-ArgList, -LineColumn)
%
%     DCG rule which parses a non-empty argument list of a compound
%     term and returns it as a real Prolog list.
%

argument_list(ArgList, [Line, Column]) -->
  % First argument
  expression(FirstArg, [Line, Column]),
  (
    % Remaining arguments
    [comma(_, _)], argument_list(RemainingArgs, _)
  ;
    % No more arguments
    {
      RemainingArgs = []
    }
  ),
  {
    parse_expression(FirstArg, ResultFirstArg),
    ArgList = [ResultFirstArg|RemainingArgs]
  }.


%% items(-List, -LineColumn)
%
%     DCG rule which parses and returns a non-empty Prolog list.
%

items(List, [Line, Column]) -->
  % First item
  expression(Head, [Line, Column]),
  (
    % Remaining items ([Head, ...])
    [comma(_, _)], items(MoreItems, _)
  ;
    % Tail sublist ([Head|Tail])
    [list_separator(_, _)], expression(Tail, _),
    {
      parse_expression(Tail, MoreItems)
    }
  ;
    % No more items ([Head])
    {
      MoreItems = []
    }
  ),
  {
    parse_expression(Head, FirstItem),
    List = [FirstItem|MoreItems]
  }.


%% expression(-Exp, -LineColumn)
%
%     DCG rule which parses and returns an expression which can appear
%     inside a Prolog list or an argument list of a compound term.
%

expression([Exp], [Line, Column]) -->
  operator(Exp, [Line, Column]),
  {
    Exp \== ','
  }.

expression(Exp, [Line, Column]) -->
  term(Exp, [not_allowed_ops([','])], _, [Line, Column]).


%% atom_or_linguistic_term(-Atom, -LineColumn)
%
%     DCG rule which parses and returns a Prolog atom or a Bousi-Prolog
%     linguistic term built with the '#' operator (these terms are
%     internally treated as atoms).
%

atom_or_linguistic_term(Atom, [Line, Column]) -->
  atom_(Atom, [Line, Column]).

atom_or_linguistic_term(LingTerm, [Line, Column]) -->
  bpl_linguistic_term(LingTerm, [Line, Column]).



%% atom_(-Atom, -LineColumn)
%
%     DCG rule which parses and returns a Prolog atom. Note that 'atom'
%     is a reserved predicate and thus cannot be redefined; that's why
%     this rule is called 'atom_' instead.
%

atom_(Atom, [Line, Column]) -->
  [name(Atom, [Line, Column])].

atom_([], [Line, Column]) -->
  [left_bracket(_, [Line, Column])], [right_bracket(_, _)].


%% operator(-Operator, -LineColumn)
%
%     DCG rule which parses and returns a Prolog predefined operator
%     or a custom operator declared with an op/3 directive.
%

operator(Operator, [Line, Column]) -->
  [name(Operator, [Line, Column])],
  {
    operator_type(_, _, Operator)
  }.

operator(Operator, [Line, Column]) -->
  [comma(Operator, [Line, Column])].


%% symbol(-Symbol, -LineColumn)
%
%     DCG rule which parses and returns a symbol that can appear in
%     a Bousi-Prolog equation, i.e., an atom or a linguistic term
%     built with the '#' operator.
%

symbol(Symbol, [Line, Column]) -->
  [name(Symbol, [Line, Column])].

symbol(Symbol, [Line, Column]) -->
  bpl_linguistic_term(Symbol, [Line, Column]).


%% integer_number(-Number, -LineColumn)
%
%     DCG rule which parses and returns a positive or negative integer
%     number.
%

integer_number(Number, [Line, Column]) -->
  [integer(Number, [Line, Column])].

integer_number(Number, [Line, Column]) -->
  [name(Sign, [Line, Column])], [integer(UnsignedNumber, _)],
  {
    (Sign == '-' ->
      Number is -UnsignedNumber
    ; (Sign == '+' ->
      Number is UnsignedNumber
    ;
      fail
    ))
  }.


%% float_number(-Number, -LineColumn)
%
%     DCG rule which parses and returns a positive or negative floating
%     point number.
%

float_number(Number, [Line, Column]) -->
  [float(Number, [Line, Column])].

float_number(Number, [Line, Column]) -->
  [name(Sign, [Line, Column])], [float(UnsignedNumber, _)],
  {
    (Sign == '-' ->
      Number is -UnsignedNumber
    ; (Sign == '+' ->
      Number is UnsignedNumber
    ;
      fail
    ))
  }.


%% number_(-Number, -LineColumn)
%
%     DCG rule which parses and returns a positive or negative integer or
%     floating point number. Note that 'number' is a reserved predicate
%     and thus cannot be redefined; that's why this rule is called
%     'number_' instead.
%

number_(Number, [Line, Column]) -->
  integer_number(Number, [Line, Column]).

number_(Number, [Line, Column]) -->
  float_number(Number, [Line, Column]).


%% number_or_variable(-NumberOrVariable, -LineColumn)
%
%     DCG rule which parses and returns an integer number, a floating
%     point number, or a variable.
%

number_or_variable(Number, [Line, Column]) -->
  number_(Number, [Line, Column]).

number_or_variable(Variable, [Line, Column]) -->
  [variable(Variable, [Line, Column])].



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DCG (Definite Clause Grammar) additional rules for parsing queries
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% query(-Query, -NewHypoRules)
%
%     DCG rule which parses and returns a Prolog/Bousi-Prolog query,
%     which must be a term optionally ending with a period in which the
%     main operator isn't ':-'.
%
%     If the query contains assumptions, they are returned in 
%     NewHypoRules.
%

query(Query, NewHypoRules) -->
  %term(Term, [not_allowed_ops([':-'])], NewHypoRules1, _),
  term(Term, [], NewHypoRules1, _),
  query_end,
  {
    parse_expression(Term, ResultTermWithoutCtrsAndDegreeVars),
    build_query(ResultTermWithoutCtrsAndDegreeVars, Query, NewHypoRules2),
    append(NewHypoRules1, NewHypoRules2, NewHypoRules)
  }.


%% build_query(+TermWithoutCtrsAndDegreeVars, -Query, -NewHypoRules)
%
%     Adds to a term the degree variable, and also,
%     for algorithms a2 and a3, input and output 
%     constraint variables.
%
%     If the query contains assumptions, they are returned in 
%     NewHypoRules.
%

build_query(TermWithoutCtrsAndDegreeVars, Query, NewHypoRules) :-
  build_goal(TermWithoutCtrsAndDegreeVars, Query, NewHypoRules),
  Query = [_ResultTerm, BodyConstraintBlockVars, _BodyDegreeVars],
  init_ctr_store(BodyConstraintBlockVars),
  % No need to initialize DegreeVars (a variable is accepted as a degree)
  !.

    
%% build_goal(+TermWithoutCtrsAndDegreeVars, -Query, -NewHypoRules)
%
%     Helper predicate for build_query/4.
%

% WARNING: Revise hypo(query, [], [])
build_goal(TermWithoutCtrsAndDegreeVars, Goal, NewHypoRules) :-
  parsed_term_variables(TermWithoutCtrsAndDegreeVars, _RuleVars),
  translate_rule_body(TermWithoutCtrsAndDegreeVars, ResultTerm, BodyDegreeVars, BodyConstraintBlockVars, hypo(query, [], []), NewHypoRules),
  Goal = [ResultTerm, BodyConstraintBlockVars, BodyDegreeVars],
  !.


%% init_ctr_store(?Constraints)
%
%     Initializes the input constraint store for algorithms 'a2' and 'a3'.
%     It does nothing for others.
%

init_ctr_store([Cin|_]) :-
  flags:get_bpl_flag(weak_unification('a3')),
  !,
  empty_assoc(Cin).
  
init_ctr_store([Cin|_]) :-
  flags:get_bpl_flag(weak_unification('a2')),
  !,
  Cin=[].
  
init_ctr_store(_Cin).


%% query_end//
%
%     DCG rule which parses the end of a Prolog/Bousi-Prolog query,
%     which can be a period or just the end-of-file character.
%

query_end -->
  [name('.', _)], [eof(_, _)].

query_end -->
  [eof(_, _)].



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DCG (Definite Clause Grammar) additional rules for managing errors
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% bousi_prolog_program_error(-DirectivesDiffList 
%%                            -ReplacedDirectivesDiffList
%%                            -RulesDiffList
%%                            -EquationsDiffList)
%
%     DCG rule which parses a Prolog statement with syntax errors,
%     saves error in messages/1 predicate, looks for the beginning of
%     the next statement (actually, looks for the next period '.' in
%     source code) and then goes on parsing using
%     bousi_prolog_program/6 rule.
%
%     @see bousi_prolog_program/6
%

bousi_prolog_program_error(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Syntax error in rule body
  basic_prolog_term(_Head, _, _), [name(':-', _)],
  term_error([not_allowed_ops([':-'])], _),
  next_statement([LineError, ColumnError]),
  {
    !,
    add_message('Syntax error in rule body.', LineError, ColumnError, error)
  },
  bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo).

bousi_prolog_program_error(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Syntax error in directive body
  [name(':-', _)], term_error([not_allowed_ops([':-'])], _),
  next_statement([LineError, ColumnError]),
  {
    !,
    add_message('Syntax error in directive.', LineError, ColumnError, error)
  },
  bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo).

bousi_prolog_program_error(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Syntax error in fact or rule head
  term_error([not_allowed_ops([':-'])], _),
  next_statement([LineError, ColumnError]),
  {
    !,
    add_message('Syntax error in fact or rule head.', LineError, ColumnError, error)
  },
  bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo).

bousi_prolog_program_error(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo) -->
  % Unknown syntax error in statement
  next_statement([LineError, ColumnError]),
  {
    !,
    add_message('Syntax error in statement.', LineError, ColumnError, error)
  },
  bousi_prolog_program(Diri, Diro, RDiri, RDiro, Ruli, Rulo, Equi, Equo).

bousi_prolog_program_error(Dir, Dir, RDir, RDir, Rul, Rul, Equ, Equ) -->
  % End-of-file (this rule is executed only when 'eof' token has been
  % taken from input by next_statement/1 in one of the previous rules)
  [].


%% term_error(+Options, -LineColumn)
%
%     DCG rule which parses a partial Prolog/Bousi-Prolog term, until a
%     syntax error is found.
%
%     Options is a list that can only contain the following item:
%
%      * not_allowed_ops(+Operators): Operators is a list of operators
%        that can't be used as the main operator in this term.
%
%     @see term/3
%

%% remaining_term_error(+Options)
%
%     Helper DCG rule used to reduce backtracking and increase
%     analysis speed.
%

term_error(Options, [Line, Column]) -->
  % Term with prefix operator (must appear before following rule)
  operator(Operator, [Line, Column]),
  {
    (member(not_allowed_ops(NotAllowedOps), Options) ->
      not(member(Operator, NotAllowedOps))
    ;
      true
    ),
    operator_type(_, OpType, Operator),
    member(OpType, [fx, fy])
  },
  term_error(Options, _).

term_error(_, [Line, Column]) -->
  % Compound term (functional notation)
  atom_(_, [Line, Column]), [left_parenthesis(_, _)], term_error([], _).

term_error(Options, [Line, Column]) -->
  % Term with infix operator, or without operator
  basic_term(_, [], _, [Line, Column]), remaining_term_error(Options).

term_error(_, [Line, Column]) -->
  % Term in parenthesis
  [left_parenthesis(_, [Line, Column])], term_error([], _).

term_error(_, [_Line, _Column]) -->
  % Can't go on parsing term, syntax error is around here
  [].


remaining_term_error(Options) -->
  % Infix operator
  operator(Operator, _),
  {
    (member(not_allowed_ops(NotAllowedOps), Options) -> 
      not(member(Operator, NotAllowedOps))
    ;
      true
    ),
    operator_type(_, OpType, Operator),
    member(OpType, [xfx, xfy, yfx])
  },
  term_error(Options, _).

remaining_term_error(_) -->
  % No operator
  [].


%% next_statement(-LineColumn)
%
%     DCG rule which matches all tokens in input until end-of-file is
%     reached or a period is found. The end-of-file or period token
%     is also taken from input. This rule is used by
%     bousi_prolog_program_error/6 to skip a statement with syntax
%     errors.
%

next_statement([Line, Column]) -->
  % There is a period before end-of-file
  all_tokens('.', [LineError, ColumnError]), [name('.', [LinePeriod, ColumnPeriod])],
  {
    ((var(LineError), var(ColumnError)) ->
      % There aren't any tokens before period
      Line = LinePeriod, Column = ColumnPeriod
    ;
      % There is at least one token before period
      Line = LineError, Column = ColumnError
    )
  }.

next_statement([Line, Column]) -->
  % There are no periods before end-of-file
  all_tokens('.', [LineError, ColumnError]), [eof(_, [LineEOF, ColumnEOF])],
  {
    ((var(LineError), var(ColumnError)) ->
      % There aren't any tokens before end-of-file
      Line = LineEOF, Column = ColumnEOF
    ;
      % There is at least one token before end-of-file
      Line = LineError, Column = ColumnError
    )
  }.


%% all_tokens(+EndText, -LineColumn)
%
%     DCG rule which matches all tokens in input until end-of-file is
%     reached or a token with text EndText is found. The end-of-file
%     or EndText token won't be taken from input, so this rule can
%     succeed without eating any token.
%

all_tokens(EndText, [Line, Column]) -->
  [Token],
  {
    Token =.. [Type, Text, [Line, Column]],
    Type \== eof, Text \== EndText
  },
  all_tokens(EndText, _).

all_tokens(_, [_Line, _Column]) -->
  [].



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for parsing expressions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% parse_expression(+Expression, ?SyntaxTree)
%
%     Scans a "plain" expression list returned by term/2 DCG rule and
%     returns the syntax tree which represents the real expression
%     structure. If an operator priority clash is detected, an error
%     will be generated and SyntaxTree will be unified with 'error'.
%

%% parse_expression_aux(+Expression, ?SyntaxTree)
%
%     Helper predicate which doesn't catch exceptions.
%
%     @throws parse_error(Message, [Line, Column]) An operator
%      priority clash was detected.
%
%     @see parse_expression/2
%


parse_expression(Expression, SyntaxTree) :-
  catch((
    parse_expression_aux(Expression, SyntaxTree)
  % (catcher)
  ), parse_error(Message, [Line, Column]), (
    % Operator priority clash (here SyntaxTree can be unified
    % with any term, as no TPL file nor query won't be generated)
    (add_message(Message, Line, Column, error), SyntaxTree = error)
  )), !.
  

parse_expression_aux([Expression], Expression) :-
  % The expression just contains a single operand
  !.

parse_expression_aux(Expression, SyntaxTree) :-
  % Looks for the highest operator in Expression
  highest_operator(Expression, [Operator, _, _, Index]),
  !,
  % Splits the "plain" expression list into two sublists
  length(LeftExpression, Index),
  append(LeftExpression, [exp_operator(Operator, Location, _)|RightExpression], Expression), !,
  (Location == infix ->
    % Builds the syntax tree
    SyntaxTree =.. [Operator, LeftSyntaxTree, RightSyntaxTree],
    % Parses each sublist separately
    parse_expression_aux(LeftExpression, LeftSyntaxTree), !,
    parse_expression_aux(RightExpression, RightSyntaxTree)
  ;
    % Prefix operator
    % Parses each sublist separately
    % (for prefix operators there's just 1 sublist)
    LeftExpression = [],
    parse_expression_aux(RightExpression, SingleSyntaxTree), !,
    % Builds the syntax tree
    (((Operator == '-' ; Operator == '+'), number(SingleSyntaxTree)) ->
      % This special treatment is needed in order to keep positive
      % and negative numbers as atomic (non-compound) terms, which is
      % the standard behavior according to ISO Prolog
      (Operator == '-' ->
        SyntaxTree is -SingleSyntaxTree
      ;
        SyntaxTree is SingleSyntaxTree
      )
    ;
      SyntaxTree =.. [Operator, SingleSyntaxTree]
    )
  ).


%% highest_operator(+Expression, ?OpInfo)
%
%     Scans a "plain" expression list returned by term/3 DCG rule and
%     extracts the main operator, i.e., the highest priority operator
%     that should be placed as the root node of the syntax tree.
%     OpInfo will be unified with a list containing operator symbol,
%     precedence, type and index (related to Expression list).
%
%     @throws parse_error(Message, LineColumn) An operator
%      priority clash was detected.
%

%% highest_operator_aux(+Expression, +Index, +OpInfo, -FinalOpInfo)
%
%     Helper predicate used by highest_operator/2. Index specifies
%     the position of the first element in Expression list; FinalOpInfo
%     is unified with OpInfo when Expression list has no more items.
%
%     @throws parse_error(Message, LineColumn) An operator
%      priority clash was detected.
%
%     @see highest_operator/2
%

highest_operator(Expression, OpInfo) :-
  highest_operator_aux(Expression, 0, [_, -1, _, _], OpInfo).


highest_operator_aux([], _, OpInfo, OpInfo).

highest_operator_aux([Head|Tail], Index, OpInfo, FinalOpInfo) :-
  Head \= exp_operator(_, _, _),
  !,
  % If head item isn't an operator, go on analyzing the expression list
  NewIndex is Index + 1,
  highest_operator_aux(Tail, NewIndex, OpInfo, FinalOpInfo).

highest_operator_aux([Head|Tail], Index, [HighOp, HighPrio, HighType, HighPos], FinalOpInfo) :-
  nonvar(Head),
  Head = exp_operator(Op, Location, [LineOp, ColumnOp]),
  NewIndex is Index + 1,
  % Gets the type and precedence of the operator
  operator_type(Prio, Type, Op),
  (Location == infix ->
    member(Type, [xfx, xfy, yfx])
  ;
    member(Type, [fx, fy])
  ),
  % Checks if the operator is a '+' or '-' and the next term is a number
  (((Op == '-' ; Op == '+'), Type == fy,
    Tail = [Number|_], number(Number)) ->
    % This special treatment is needed in order to ignore the '+' and
    % '-' operators that indicates the sign of a number, because
    % positive and negative numbers must be atomic (non-compound) terms
    % according to the ISO Prolog standard
    highest_operator_aux(Tail, NewIndex, [HighOp, HighPrio, HighType, HighPos],
                         FinalOpInfo)
  ;
    (Prio > HighPrio ->
      % A new highest operator was found
      highest_operator_aux(Tail, NewIndex, [Op, Prio, Type, Index], FinalOpInfo)
    ;
    (Prio =:= HighPrio ->
      % If an expression has several operators with the same priority,
      % priority clash will happen if the type of one of the operators
      % ends with "x" and the next one starts with "x" or "f"
      % Examples: - xfy xfy fx yfx yfx -> OK
      %           - yfx yfx xfy xfy fx -> Priority clash (yfx xfy)
      %           - xfx yfx fy xfx fx -> Priority clash (yfx fy)
      (Type == yfx ->
        % Left associative, change highest operator
        highest_operator_aux(Tail, NewIndex, [Op, Prio, Type, Index], FinalOpInfo)
      ;
      (((Type == xfy ; Type == xfx ; Type == fx ; Type == fy),
        (HighType == xfy ; HighType == fy)) ->
        % Right associative, don't change highest operator
        highest_operator_aux(Tail, NewIndex, [HighOp, HighPrio, HighType, HighPos],
                             FinalOpInfo)
      ;
        % Operators can't be associated
        swritef(ErrorMessage, 'Operator priority clash caused by \'%w\'.', [Op]),
        throw(parse_error(ErrorMessage, [LineOp, ColumnOp]))
      ))
    ;
      % Keeps the current highest operator
      highest_operator_aux(Tail, NewIndex, [HighOp, HighPrio, HighType, HighPos],
                           FinalOpInfo)
    ))
  ).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates related with errors and warnings
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% add_message(+Message, +Line, +Column, +Type)
%
%     Adds a new message to the list stored in messages/1. Error
%     messages should pass Type = 'error', whereas warning messages
%     should pass Type = 'warning'.
%

%% add_message(+Message, +Line, +Type)
%
%     Adds a new message to the list stored in messages/1. Error
%     messages should pass Type = 'error', whereas warning messages
%     should pass Type = 'warning'.
%

add_message(Message, Line, Column, Type) :-
  % Creates the text of the message
  bpl_filename(InputFile),
  upcase_atom(Type, UpperType),
  (InputFile == '' ->
    % InputFile should only be '' if a query is being parsed
    swritef(StringMessage, '%w: %w',
            [UpperType, Message])
  ;
    swritef(StringMessage, '%w: %w:%w:%w: %w',
            [UpperType, InputFile, Line, Column, Message])
  ),
  string_to_atom(StringMessage, TextMessage),
  % Updates the message list
  NewMessage = [InputFile, Line, Column, TextMessage, Type],
  add_message(NewMessage).


add_message(Message, Line, Type) :-
  % Creates the text of the message
  bpl_filename(InputFile),
  upcase_atom(Type, UpperType),
  (InputFile == '' ->
    % InputFile should only be '' if a query is being parsed
    swritef(StringMessage, '%w: %w',
            [UpperType, Message])
  ;
    swritef(StringMessage, '%w: %w:%w: %w',
            [UpperType, InputFile, Line, Message])
  ),
  string_to_atom(StringMessage, TextMessage),
  % Updates the message list
  NewMessage = [InputFile, Line, 0, TextMessage, Type],
  add_message(NewMessage).


%% add_message_in_file(+File, +Message, +Type)
%
%     Adds a new message to the list stored in messages/1, using
%     a different filename than the one kept in bpl_filename/1. Error
%     messages should pass Type = 'error', whereas warning messages
%     should pass Type = 'warning'.
%

add_message_in_file(File, Message, Type) :-
  % Creates the text of the message
  upcase_atom(Type, UpperType),
  swritef(StringMessage, '%w: %w: %w', [UpperType, File, Message]),
  string_to_atom(StringMessage, TextMessage),
  % Updates the message list
  NewMessage = [File, 0, 0, TextMessage, Type],
  add_message(NewMessage).


%% add_message(+Message)
%
%     Adds a new message to the list of messages stored in messages/1
%     dynamic predicate.
%

add_message(Message) :-
  messages(CurrentMessages),
  retract(messages(CurrentMessages)),
  assert(messages([Message|CurrentMessages])).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for processing goals and rules
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% add_degree_variables(+Term, -Result, -DegreeVars)
%
%     Scans Term, adds a degree variable to all the calls to
%     non-predefined predicates and returns the resulting term in
%     Result. DegreeVars will be unified with a list containing the
%     degree variables appended by this predicate. See
%     degree_variable_scanner/4 for an example.
%
%     @see degree_variable_scanner/4
%

%% add_degree_variables(+Term, -Result, -DegreeVars)
%
%     Helper predicate for add_degree_variables/3.
%

% Weak unification algorithm 'a1'
add_degree_variables(Term, Result, DegreeVars) :-
  flags:get_bpl_flag(degree_accumulator(false)),
  !,
  add_degree_variables_aux(Term, Result, DegreeVars).

% Weak unification algorithm 'a2' or 'a3'
% Optimized version: With degree accumulator
add_degree_variables(Term, Result, DegreeVars) :-
  flags:get_bpl_flag(disjunction(true)),
  !,
  % Add degree variables to Term
  add_degree_variables_aux(Term, NewResult, DegreeVars),
  % Extract each disjunct in NewResult
  disjunctive_normal_form(NewResult, DisjTerm),
  utilities:disjunction_to_list(DisjTerm, Disjuncts),
  % Extract the degree variables in each disjunct
  maplist(extract_degree_variables, Disjuncts, RestDegreeVars),
  % Chain the degree variables of each disjunct
  maplist(link_accumulator_variables, RestDegreeVars),
  utilities:cond_filter_element(utilities:var_or_empty_list, [DegreeVars|RestDegreeVars], NonEmptyDegreeVarsList),
  % Link heads and last element of the degree variables lists
  % For example: [[X1, X2, X3], [Y1, Y2]] -> X1 = Y1, X3 = Y2
  link_heads_and_lasts(NonEmptyDegreeVarsList),
  % Recover the original form of built-in predicates, along 
  % with the call Din = Dout
  replace_builtin(NewResult, Result).

% Weak unification algorithm 'a2' or 'a3'
% Non-optimized version: Without degree accumulator
add_degree_variables(Term, Result, DegreeVars) :-
  % Add degree variables to Term
  add_degree_variables_aux(Term, Result, DegreeVars),
  link_accumulator_variables(DegreeVars).


add_degree_variables_aux(Term, Result, DegreeVars) :-
  utilities:process_term(Term, Result, 
                         [parser:degree_variable_scanner],
                         [parser:control_predicate_tester],
                         [], DegreeVars).


%% register_disjunction_in_term(+Term)
%
%    Registers if a disjunction is present in the term.
%    Sets the BPL flag disjunction to true if a disjunction
%    is in the term.
%

register_disjunction_in_term(Term) :-
  ((member_term_with_pattern(_Subterm, Term, [T, T = ';'(_, _)]) ;
    member_term_with_pattern(_Subterm, Term, [T, T = catch(_, _, _)]))
   -> flags:set_bpl_flag(disjunction(true))
   ;  flags:set_bpl_flag(disjunction(false))).


%% extract_degree_variables(+Term, -LastTwoArgsList)
%
%    Given a conjunctive term, returns a list of the last two arguments
%    of each conjunct.
%

extract_degree_variables(Term, DegreeVars) :-
  utilities:process_term(Term, Term, 
                         [parser:extract_degree_variables_scanner],
                         [parser:control_predicate_tester],
                         [], DegreeVars).


%% link_accumulator_variables(+Vars)
%
%    Links pairs of successive variables in a list of 
%      even length, starting from the second variable.
%

%% link_pairs(+Current, +Rest)
%
%    Helper predicate to link pairs of successive variables.
%

link_accumulator_variables(Var) :-
  var(Var),
  !. % Base case: No accumulator variables

link_accumulator_variables([]) :- !. % Base case: No accumulator variables

link_accumulator_variables([_, _]) :- !. % Base case: two-element list, nothing to link

link_accumulator_variables([_, Second|Rest]) :-
    link_pairs(Second, Rest).


link_pairs(_, []) :- !. % Base case: no more elements to link

link_pairs(X, [X, Y|Rest]) :-
    link_pairs(Y, Rest).


%% link_heads_and_lasts(+ListOfLists)
%
%     Predicate to unify all the heads and all the last elements 
%     of a list of lists
%

%% unify_head(Head, List)
%
%     Helper predicate to unify the head with the given head


link_heads_and_lasts([]).

link_heads_and_lasts([Var|Lists]) :-
  var(Var),
  !,
  link_heads_and_lasts(Lists).

link_heads_and_lasts([[H|T]|Lists]) :-
  % Ensure all heads are equal
    maplist(unify_head(H), Lists),
    % Ensure all last elements are equal
    maplist(last_element(_Last), [[H|T]|Lists]),
    !.


unify_head(H, [H|_]).


% Helper predicate to extract the last element of a list
last_element(L, List) :-
    last(List, L).



%% link_ho_goal(+Term, +ConstraintsVars)
%
%     Links the constraint variables of a higher-order call
%     to the constraint variables of its goal if its is
%     also a HO call.
%

link_ho_goal(Term, ConstraintsVars) :-
  Term =.. [Functor|Args],
  Functor == ho_call,
  Args = [_HOList, _Compiled, _Goal, ConstraintsVars|_],
  !.

link_ho_goal(_Term, _ConstraintsVars).


%% add_constraint_variables(+Term, -Result, -ConstraintsVars)
%
%     If the selected weak unification algorithm is either 'a2' or 'a3', 
%     then scans Term, adds an input and output constraint variables
%     to all the calls to non-predefined predicates and returns the 
%     resulting term in Result. ConstraintsVars will be unified
%     with a list containing the constraint variables appended 
%     by this predicate. See
%     constraint_variables_scanner/4 for an example.
%     For other weak unification algorithms, Result is the same Term 
%     and ConstraintsVars is left as is.
%
%     If Term contains a disjunction, all possible paths are examined
%     to link both each input and output constraint variables 
%     of each disjunct. Each disjunct is processed separately and the
%     resulting constraint variables are linked together.
%     The predicate disjunctive_normal_form/2 is used to extract each
%     disjunct.
%
%     @see constraint_variables_scanner/4
%

%% add_constraint_variables_aux(+Term, -Result, -ConstraintsVars)
%
%     Helper predicate for add_constraint_variables/3.

% Weak unification algorithm 'a1'
add_constraint_variables(Term, Term, _ConstraintsVars) :-
  flags:get_bpl_flag(weak_unification('a1')),
  !.

% Weak unification algorithm 'a2' or 'a3'
% Optimized version: With degree accumulator
add_constraint_variables(Term, Result, ConstraintsVars) :-
  flags:get_bpl_flag(disjunction(true)),
  !,
  % Add constraint variables to Term
  add_constraint_variables_aux(Term, NewResult, ConstraintsVars),
  % Extract each disjunct in NewResult
  disjunctive_normal_form(NewResult, DisjTerm),
  utilities:disjunction_to_list(DisjTerm, Disjuncts),
  % Extract the constraint variables in each disjunct
  maplist(extract_constraint_variables, Disjuncts, RestBlockCtrVars),
  % Chain the constraint variables of each disjunct
  maplist(link_accumulator_variables, RestBlockCtrVars),
  utilities:cond_filter_element(utilities:var_or_empty_list, [ConstraintsVars|RestBlockCtrVars], NonEmptyVarsList),
  % Link heads and last element of the degree variables lists
  % For example: [[X1, X2, X3], [Y1, Y2]] -> X1 = Y1, X3 = Y2
  link_heads_and_lasts(NonEmptyVarsList),
  % Recover the original form of built-in predicates, along 
  % with the call Cin = Cout
  replace_builtin(NewResult, Result).
  
% Weak unification algorithm 'a2' or 'a3'
% Non-optimized version: Without degree accumulator
add_constraint_variables(Term, Result, ConstraintsVars) :-
  add_constraint_variables_aux(Term, Result, ConstraintsVars),
  link_accumulator_variables(ConstraintsVars).


add_constraint_variables_aux(Term, Result, ConstraintsVars) :-
  utilities:process_term(Term, Result,
                         [parser:constraint_variables_scanner],
                         [parser:control_predicate_tester],
                         [], ConstraintsVars).



%% extract_constraint_variables(+ConjunctiveTerm, -ConstraintsVars)
%
%    Given a conjunctive term, returns a list of the constraint
%    variables of each conjunct.
%

extract_constraint_variables(Term, ConstraintsVars) :-
  utilities:process_term(Term, Term, 
                         [parser:extract_constraint_variables_scanner],
                         [parser:control_predicate_tester],
                         [], ConstraintsVars).
  

%% add_hypothetical_term(+Term, -Result, ?HypoData, -NewHypoRules)
%
%     When hypothetical reasoning is enabled, this predicate
%     adds the hypothetical term to rule's head for passing shared 
%     variables, rule identifier, rule context and current context.
%     This term must be just after any head variable (if any) and
%     before constraint variables and degree variables.
%
%     HypoData = hypo(head, HypoTerm)
%              | hypo(body, RuleVars, CD)

add_hypothetical_term(Term, Term, _HypoData, []) :-
  flags:get_bpl_flag(hypothetical(false)),
  !.

add_hypothetical_term(Term, Result, HypoData, HypoRules) :-
  utilities:process_term(Term, Result,
                         [parser:hypothetical_term_scanner],
                         [parser:control_predicate_tester],
                         [HypoData, []], % [HypoData, InRules] 
                         [_, HypoRules]).


%% check_singleton(+Term, +Line)
%
%     Scans Term and looks for singleton variables (i.e., variables
%     which are used only once) and singleton-marked variables (i.e.,
%     variables which start with an underscore character). If at least
%     one singleton variable is found, a warning will be generated
%     using the specified Line. Besides, if at least one singleton-marked
%     variable is used more than once, another warning will be generated.
%     Note that, in this context, "variable" means "atom starting with
%     an uppercase letter or an underscore character".
%

check_singleton(Term, Line) :-
  % Extracts and counts all the variables of Term
  utilities:process_term(Term, _Result,
                         [parser:get_variables_scanner],
                         [],
                         [], AllVars),
  % Gets the name of the variables that have been found only once,
  % ignoring the singleton-marked variables (those starting with '_')
  findall(Var, (member([Var, Count], AllVars),
                atom_chars(Var, [First|_]), First \== '_',
                (Count =:= 1)), SingletonVars),
  (SingletonVars \== [] ->
    % One or more singleton variables were found
    swritef(MessageSV, 'Singleton variables: %w', [SingletonVars]),
    add_message(MessageSV, Line, warning)
  ;
    % No singleton variables were found
    true
  ),
  % Gets the name of the singleton-marked variables (those starting with
  % '_' that aren't '_' itself) that have been found more than once
  findall(Var, (member([Var, Count], AllVars), Var \== '_',
                atom_chars(Var, [First|_]), First == '_',
                (Count > 1)), SingletonMarkedVars),
  (SingletonMarkedVars \== [] ->
    % One or more singleton-marked variables were found several times
    swritef(MessageSMV, 'Singleton-marked variables appearing more than \c
                         once: %w', [SingletonMarkedVars]),
    add_message(MessageSMV, Line, warning)
  ;
    % No repeated singleton-marked variables were found
    true
  ).


%% replace_variable_calls(+Term, -NewTerm)
%
%     Replace each variable X in a goal position by bpl_call(X)
%     to allow for higher-order variable calls.
%

replace_variable_calls(Term, Term) :-
  (var(Term) ; atomic(Term)),
  !.

replace_variable_calls(Head :- Body, Head :- NewBody) :-
  % If Term is a clause, only its body is scanned
  !,
  replace_variable_calls(Body, NewBody).

replace_variable_calls(Term, NewTerm) :-
  !,
  % Extracts and counts all the variables of Term
  % (non-control compound terms are skipped)
  utilities:process_term(Term, NewTerm,
                         [parser:replace_variable_calls_scanner],
                         [parser:control_predicate_tester],
                         _, _).

  
  
%% degree_variable_scanner(+Term, -Result, +InData, -OutData)
%
%     Scanner that can be used with process_term/6 to add degree
%     variables to each call to a user-defined or higher-order
%     predicate and get the new degree variables in OutData. For
%     every Term that is a compound term with a non-predefined
%     functor, Result will be unified with a new term with the
%     following components:
%
%      * Functor: concatenation of the prefix stored in
%        program_prefix/1, an underscore character '_' and the
%        original functor.
%      * Arguments: the ones in the original term plus
%        - A new free variable which will store the approximation
%          degree (degree accumulator optimization disabled).
%        - Two new free variables for input and output degrees
%          (degree accumulator optimization enabled).
%
%     For example, given Term = 'pred(X, 1)', this predicate will
%     return Result = 'prog_pred(X, 1, DegreeVar)' (if degree
%     accumulator optimization is disabled and program_prefix/1
%     contains 'prog') and OutData = [DegreeVar|InData].
%
%     If Term is a call to one of the supported predefined higher-
%     order predicates (for example, apply(pred, List)), Term is
%     encapsulated in the Result term ho_call, where its arguments
%     include the call, arguments, goal degree arguments and output
%     degree.
%
%     If Term is a call to a predefined predicate (for example,
%     'append([1], X, [1, 2])'), and degree accumulator optimization
%     is disabled, Result and OutData will be unified with Term and
%     InData, respectively. 
%     When degree accumulator optimization is
%     enabled, Result is the term '$builtin'((Term, Din = Dout),
%     [Din, Dout]), where Din and Dout are the input and output degree
%     variables, respectively, which will be unified at run-time by
%     Din = Dout. These degree variables must be made explicit for a
%     correct chaining of the degree variables in the body of the
%     program when disjunctions are present. A term is known to
%     include disjunctions if the BPL flag disjunction has been set to
%     true (while not declarative, this trick simplifies the current
%     implementation).
%
%     @see process_term/6
%

degree_variable_scanner(Term, NewTerm, InDegreeVars, OutDegreeVars) :-
  nonvar(Term),
  not(number(Term)),
  !,
  Term =.. [Functor|Args],
  (
    % Term represents a variable, so it can be ignored
    utilities:atom_is_variable(Functor),
    !,
    NewTerm = Term,
    OutDegreeVars = InDegreeVars
  ; % apply/2, maplist/2
    ( Functor == bpl_apply
    ; Functor == bpl_maplist
    ),
    % Args = [PartialGoal, _],
    % !,
    % ho_prepare_degree_variables([Functor|Args], no, PartialGoal, _GoalDegreeVars, NewTerm, InDegreeVars, OutDegreeVars)
    Args = [PartialGoal|MoreArgs],
    !,
    ho_prepare_degree_variables([Functor|MoreArgs], no, PartialGoal, _GoalDegreeVars, NewTerm, InDegreeVars, OutDegreeVars)
  ; % Higher order call with one goal
    ( % Negation
      Term = eval_negation(Neg, Goal),
      !,
      HOList = [eval_negation, Neg]
     ; % Call
      (Term = bpl_call(Goal) 
      ; 
       higher_order_predicate(Functor, [goal]),
       Args = [Goal]),
      !,
      HOList = [Functor] 
     ; % truth_degree
      higher_order_predicate(Functor, [goal, nongoal]),
      Args = [Goal, Number],
      !,
      HOList = [Functor, Number]
     ; % findall/3
      Functor == findall,
      Args = [Template, Goal, Bag],
      !,
      HOList = [Functor, Template, Bag]
     ; % setof/3, bagof/3
      higher_order_predicate(Functor, [nongoal, goal, nongoal]),
      Args = [Template, QuantifiedGoal, Bag],
      !,
      unbind_term(QuantifiedGoal, Goal, Bindings),
      HOList = [Functor, Template, Bag, Bindings]
    ),
    !,
    add_degree_variables_if_total_goal(Goal, NewGoal, Compiled, GoalDegreeVars),
    ho_prepare_degree_variables(HOList, Compiled, NewGoal, GoalDegreeVars, NewTerm, InDegreeVars, OutDegreeVars)
  ; % Higher order call with two goals
    ( % forall/2
      higher_order_predicate(Functor, [goal, goal]),
      Args = [Goal1, Goal2],
      !,
      Goals = (Goal1, Goal2),
      HOList = [Functor]
    ),
    !,
    add_degree_variables_if_total_goal(Goals, NewGoals, Compiled, GoalDegreeVars),
    ho_prepare_degree_variables(HOList, Compiled, NewGoals, GoalDegreeVars, NewTerm, InDegreeVars, OutDegreeVars)
  ; % Higher order call with three goals
    ( % catch/3
      higher_order_predicate(Functor, [goal, nongoal, goal]),
      Args = [Goal1, Catcher, Goal2],
      !,
      Goals = (Goal1 ; Goal2),
      HOList = [Functor, Catcher]
    ),
    !,
    add_degree_variables_if_total_goal(Goals, NewGoals, Compiled, GoalDegreeVars),
    ho_prepare_degree_variables(HOList, Compiled, NewGoals, GoalDegreeVars, NewTerm, InDegreeVars, OutDegreeVars)
  ;
    % Hypothetical operator
    Term = '=>'(HypoRules, Goal),
    flags:get_bpl_flag(hypothetical(true)),
    !,
    NewTerm = '=>'('_$'(HypoRules), NewGoal), % Do not apply this scanner to HypoRules. It will be applied later when translating HypoRules (It should be a better way to do this)
    add_degree_variables(Goal, NewGoal, GoalDegreeVars),
    append(InDegreeVars, GoalDegreeVars, OutDegreeVars)
  ;
    % Predefined predicate (built-in) but conjunction, disjunction...
    flags:get_bpl_flag(degree_accumulator(true)),
    flags:get_bpl_flag(disjunction(true)),
    \+ is_body_connective(Term),
    utilities:builtin(Term),
    !,
    NewTerm = '$builtin'((Term, Din=Dout), [Din, Dout]),
    append(InDegreeVars, [Din, Dout], OutDegreeVars)
  ;
    % quoted term, or existential operator
    (utilities:builtin(Term) ;
     utilities:is_quoted(Functor, '\'') ;
     Functor == '^' ;
     Functor == '$builtin', flags:get_bpl_flag(degree_accumulator(true))),
    !,
    NewTerm = Term,
    OutDegreeVars = InDegreeVars
  ;
    % Defined predicates
    flags:get_bpl_flag(degree_accumulator(false)) 
    ->
      % Non-optimized version: Without degree accumulator
      % Adds the degree variable to the scanned term
      append(Args, [NewDegreeVar], NewArgs),
      prefixed_predicate_name(Functor, NewFunctor),
      NewTerm =.. [NewFunctor|NewArgs],
      % Updates the dynamic degree variable list
      OutDegreeVars = [NewDegreeVar|InDegreeVars]
    ;
      % Optimized version: With degree accumulator
      % Adds the pair of input and output degree variables
      % to the scanned term
      append(Args, [NewDegreeInVar,NewDegreeOutVar], NewArgs),
      prefixed_predicate_name(Functor, NewFunctor),
      NewTerm =.. [NewFunctor|NewArgs],
      % Updates the dynamic degree variable list
      append(InDegreeVars, [NewDegreeInVar, NewDegreeOutVar], OutDegreeVars)
  ),
  !.

degree_variable_scanner(Term, Term, InDegreeVars, InDegreeVars).
  % Term can be ignored because it's a variable or a number

  
%% add_degree_variables_if_total_goal(+Goal, -NewGoal, +Compiled, -GoalDegreeVars)
%
%     Adds degree variables to the goal if it is a total goal.
%     If the goal is a partial goal, it will be passed as is.
%     Compiled is unified with 'yes' if the goal is a total goal
%     and 'no' if it is a partial goal.
%

add_degree_variables_if_total_goal(Goal, NewGoal, Compiled, GoalDegreeVars) :-
  (is_total_goal(Goal)
   ->
    % Call with a known argument
    add_degree_variables(Goal, NewGoal, GoalDegreeVars),
    Compiled = yes
   ;
    % Call with a partial goal as argument
    NewGoal = Goal,
    Compiled = no
  ).


%% ho_prepare_degree_variables(+HOList, +Compiled, +Goal, +GoalDegreeVars, -NewTerm, -InDegreeVars, -OutDegreeVars)
%
%     Encapsulates the Goal for the higher-order predicate call 
%     included in HOList = [HOCall|HOArguments] where Goal is the 
%     argument of the higher-order call HOCALL, which may have other 
%     arguments (for example, in the crisp negation \+ Goal, the list 
%     HOList is [eval_negation, \+]).
%

% Non-optimized version: Without degree accumulator
ho_prepare_degree_variables(HOList, Compiled, Goal, GoalDegreeVars, NewTerm, InDegreeVars, OutDegreeVars) :-
  flags:get_bpl_flag(degree_accumulator(false)),
  !,
  append([ho_call, HOList], [Compiled, Goal, GoalDegreeVars, NewDegreeVar], NewTermList),
  NewTerm =.. NewTermList,
  % Updates the dynamic degree variable list
  append([NewDegreeVar], InDegreeVars, OutDegreeVars).
  
% Optimized version: With degree accumulator
ho_prepare_degree_variables(HOList, Compiled, Goal, GoalDegreeVars, NewTerm, InDegreeVars, OutDegreeVars) :-
  append([ho_call, HOList], [Compiled, Goal, GoalDegreeVars, NewDegreeInVar, NewDegreeOutVar], NewTermList),
  NewTerm =.. NewTermList,
  % Updates the dynamic degree variable list linking the last 
  % variable in InDegreeVars with the first one in the body degree
  % variable list
  append(InDegreeVars, [NewDegreeInVar, NewDegreeOutVar], OutDegreeVars).


%% extract_degree_variables_scanner(+Term, -Term, +InData, -OutData)
%
%     Scanner that can be used with process_term/6 to extract
%     degree variables in OutData from each call to a 
%     user-defined predicate. 
%
%     @see process_term/6
%

extract_degree_variables_scanner(Term, Term, InDegreeVars, OutDegreeVars) :-
    nonvar(Term),
    not(number(Term)),
    !,
    Term =.. [Functor|Args],
    (
      % Term represents a variable, so it can be ignored
      utilities:atom_is_variable(Functor),
      !,
      OutDegreeVars = InDegreeVars
    ;
      % Negation (GoalDegreeVars must not be passed to OutDegreeVars)
      Functor = ho_call,
      Args = [[eval_negation, Neg], yes, _Goal, _GoalDegreeVars|DegreeVars],
      (Neg = not ; Neg = (\+)),
      !,
      append(InDegreeVars, DegreeVars, OutDegreeVars)
    ;
      % Higher order call
      Functor = ho_call,
      Args = [HOList, Compiled, _Goal, GoalDegreeVars|DegreeVars],
      !,
      ( HOList = [HOPredicate|_],
        ho_crisp_predicate(HOPredicate)
        -> OutDegreeVars = InDegreeVars
        ; (Compiled == no
          -> append(InDegreeVars, DegreeVars, OutDegreeVars)
          ;  GoalDegreeVars = [Din|_],
             link_last_variable(GoalDegreeVars, Dout),
             append(InDegreeVars, [Din, Dout], OutDegreeVars)
          )
      )
    ;
      % Hypothetical operator
      flags:get_bpl_flag(hypothetical(true)),
      Term = '=>'(_HypoRules, _Goal),
      !,
      OutDegreeVars = InDegreeVars
    ;
      % Predefined predicate (built-in) withouth degree accumulator
      flags:get_bpl_flag(degree_accumulator(true)),
      flags:get_bpl_flag(disjunction(true)),
      Term = '$builtin'(_Builtin, [Din, Dout]),
      !,
      append(InDegreeVars, [Din, Dout], OutDegreeVars)
    ;
      % Predefined predicate (built-in).
      % quoted term, or existential operator
      ( utilities:builtin(Term) ; 
        utilities:is_quoted(Functor, '\'') ;
        Functor == '^' ),
      !,
      OutDegreeVars = InDegreeVars
    ;
      % Defined predicates
      % Optimized version 
      % Extract the degree variables from the scanned term
      append(_, [DegreeInVar,DegreeOutVar], Args),
      % Add the degree variables to the output data
      append(InDegreeVars, [DegreeInVar, DegreeOutVar], OutDegreeVars)
    ),
    !.
  
  extract_degree_variables_scanner(Term, Term, InDegreeVars, InDegreeVars).
    % Term can be ignored because it's a variable or a number


%% ho_crisp_predicate(+Predicate)
%
%   List of higher-order predicates that always return
%   an approximation degree 1.0.
%

ho_crisp_predicate(truth_degree).
ho_crisp_predicate(forall).
ho_crisp_predicate(findall).
ho_crisp_predicate(setof).
ho_crisp_predicate(bagof).


%% prefixed_predicate_name(+Functor, -NewFunctor)
%
%     Prefixes a user-defined predicate name with the name of the file.
%
%     Prolog predefined predicates that need adding constraints 
%     and degree are not prefixed.
%

prefixed_predicate_name(Functor, Functor) :-
      memberchk(Functor, [weak_unify_a1, 
                          weak_unify_a2, 
                          weak_unify_a3]),
      !.
         
prefixed_predicate_name(Functor, NewFunctor) :-
      program_prefix(ProgramName),
      concat_atom([ProgramName, '_', Functor], NewFunctor).


%% replace_builtin(+Term, -Result)
%
%     Replaces the term '$builtin'(Builtin, [Din, Dout]) in Term
%     by Builtin returning Result.
%

replace_builtin(Term, Result) :-
  flags:get_bpl_flag(disjunction(true)),
  !,
  utilities:process_term(Term, Result, 
                         [parser:replace_builtin_scanner],
                         [parser:control_predicate_tester],
                         [], []).

replace_builtin(Term, Term).


%% replace_builtin_scanner(+Term, -Result, +InData, -OutData)
%
%     Scanner that can be used with process_term/6 to extract
%     degree variables in OutData from each call to a 
%     user-defined predicate. 
%
%     @see process_term/6
%

replace_builtin_scanner(Term, Result, InVars, InVars) :-
    nonvar(Term),
    Term = '$builtin'(Builtin, _),
    !,
    Result = Builtin
    ;
    Result = Term.


%% constraint_variables_scanner(+Term, -Result, +InData, -OutData)
%
%     Scanner that can be used with process_term/6 to add an input
%     and output constraint variables to each call to a 
%     user-defined predicate, and get both the new constraint
%     variables in OutData. For every Term that is a compound term
%     with a non-predefined functor, Result will be unified with a 
%     new term with the same functor as Term and:
%
%      * Arguments: the ones in the original term plus two new free
%        variables which will store the input and output 
%        constraint stores. They are placed just before the degree.
%
%     For example, given:
%         Term   = 'prog_pred(X, 1, DegreeVar)', this 
%     predicate will return 
%         Result  = 'prog_pred(X, 1, Cin, Cout, DegreeVar)', and
%         OutData = InData ++ [Cin, Cout].
%     If constraints are passed due to a1 or a3, then:
%         Result  = 'prog_pred(X, 1, Cin, Cout, DegIn, DegOut)', and
%         OutData = InData ++ [Cin, Cout].
%
%     If Term is a call to a predefined predicate (for example,
%     'append([1], X, [1, 2])'), Result and OutData will be unified
%     with Term and InData, respectively.
%
%     @see process_term/6
%

constraint_variables_scanner(Term, NewTerm, InCtrVars, OutCtrVars) :-
  nonvar(Term),
  not(number(Term)),
  !,
  Term =.. [Functor|Args],
  (
    % Term represents a variable, so it can be ignored
    utilities:atom_is_variable(Functor),
    !,
    NewTerm = Term,
    OutCtrVars = InCtrVars
    ;
    % Higher order call
    Functor == ho_call,
    Args = [HOList, Compiled, Goal, GoalDegreeVars|TermArgs],
    ( % Negation
      HOList = [eval_negation, Neg], !,
      NewHOList = HOList
    ; % Call
      HOList = [HOPred],
      (HOPred == bpl_call 
      ;
       higher_order_predicate(HOPred, [goal])
      ), !,
      NewHOList = HOList
    ; % bpl_apply/2, maplist/2
      HOList = [HOPred, _List],
      (HOPred == bpl_apply ; HOPred == bpl_maplist), !,
      NewHOList = HOList
    ; % truth_degree/2
      HOList = [HOPred, _Number],
      higher_order_predicate(HOPred, [goal, nongoal]), !,
      NewHOList = HOList
    ; % forall/2
      HOList = [HOPred],
      higher_order_predicate(HOPred, [goal, goal]), !,
      NewHOList = HOList
    ; % catch/3
      HOList = [HOPred, _Catcher],
      higher_order_predicate(HOPred, [goal, nongoal, goal]), !,
      NewHOList = HOList
     ; % findall/3
      HOList = [HOPred, _Template, _Bag],
      higher_order_predicate(HOPred, [nongoal, goal, nongoal]), !,
      NewHOList = HOList
     ; % setof/3, bagof/3
      HOList = [HOPred, Template, Bag, Bindings],
      higher_order_predicate(HOPred, [nongoal, goal, nongoal]), !,
      append(Bindings, OutCtrVars1, NewBindings),
      NewHOList = [HOPred, Template, Bag, NewBindings]
    ),
    !,
    (Compiled == yes
     ->
      add_constraint_variables(Goal, NewGoal, OutCtrVars1),
      ( % Nothing to link nor to append when there are no 
        % constraint variables
        OutCtrVars1 == [],
        !,
        Cin = Cout,
        OutCtrVars = InCtrVars
      ; % Cout for the crisp negation must not be linked
        Neg == (\+),
        !,
        OutCtrVars1 = [Cin|_],
        link_last_variable(InCtrVars, Cin),
        %OutCtrVars = InCtrVars
        append(InCtrVars, [Cin, Cin], OutCtrVars) 
      ; % Links Cin and Cout of HO calls with the first Cin and 
        %the last variable Cout in NewGoal
        OutCtrVars1 = [Cin|_],
        link_last_variable(InCtrVars, Cin),
        link_last_variable(OutCtrVars1, Cout),
        % Updates the dynamic constraint variable list
        append(InCtrVars, OutCtrVars1, OutCtrVars) 
      ),
      NewTerm =.. [ho_call, NewHOList, Compiled, NewGoal, [Cin, Cout], GoalDegreeVars|TermArgs]
     ;
      NewTerm =.. [ho_call, HOList, Compiled, Goal, [Cin, Cout], GoalDegreeVars|TermArgs],
      append(InCtrVars, [Cin, Cout], OutCtrVars)
    )
  ;
    % Hypothetical operator
    Term = '=>'('_$'(HypoRules), Goal),
    flags:get_bpl_flag(hypothetical(true)),
    !,
    % Do not apply this scanner to HypoRules. It will be applied later
    % when translating HypoRules (It should be a better way to do this)
    NewTerm = '=>'('_$'(HypoRules), NewGoal), 
    add_constraint_variables(Goal, NewGoal, OutCtrVars1),
    append(InCtrVars, OutCtrVars1, OutCtrVars)
  ;
    % Predefined predicate (built-in) but conjunction, disjunction...
    flags:get_bpl_flag(disjunction(true)),
    \+ is_body_connective(Term),
    utilities:builtin(Term),
    !,
    NewTerm = '$builtin'((Term, Cin=Cout), [Cin, Cout]),
    append(InCtrVars, [Cin, Cout], OutCtrVars)
  ;
    % Builtin, quoted term, or existential operator
    (utilities:builtin(Term) ;
     utilities:is_quoted(Functor, '\'') ;
     Functor == '^' ;
     Functor == '$builtin', flags:get_bpl_flag(disjunction(true))),
    !,
    NewTerm = Term,
    OutCtrVars = InCtrVars
  ;
    % Defined predicates
    % Adds the constraint variables to the scanned term
    Ctrs = [_Cin, _Cout],
    (flags:get_bpl_flag(degree_accumulator(false)) 
     ->
      append(ArgsWithoutDegreeVar, [NewDegreeVar], Args),
      append(ArgsWithoutDegreeVar, Ctrs, NewArgsWithoutDegreeVar),
      append(NewArgsWithoutDegreeVar, [NewDegreeVar], NewArgs)
     ;
      append(ArgsWithoutDegreeVar, [NewDegreeInVar, NewDegreeOutVar], Args),
      append(ArgsWithoutDegreeVar, Ctrs, NewArgsWithoutDegreeVar),
      append(NewArgsWithoutDegreeVar, [NewDegreeInVar, NewDegreeOutVar], NewArgs)
    ),
    NewTerm =.. [Functor|NewArgs],
    % Updates the constraint variables list
    append(InCtrVars, Ctrs, OutCtrVars)
  ),
  !.

constraint_variables_scanner(Term, Term, InData, InData).
  % Term can be ignored because it's a variable or a number


%% extract_constraint_variables_scanner(+Term, -Term, +InData, -OutData)
%
%     Scanner that can be used with process_term/6 to extract
%     degree variables in OutData from each call to a 
%     user-defined predicate. 
%
%     @see process_term/6
%

extract_constraint_variables_scanner(Term, Term, InCtrVars, OutCtrVars) :-
  nonvar(Term),
  not(number(Term)),
  !,
  Term =.. [Functor|Args],
  (
    % Term represents a variable, so it can be ignored
    utilities:atom_is_variable(Functor),
    !,
    OutCtrVars = InCtrVars
  ;
    % Crisp negation
    Functor == ho_call,
    Args = [[eval_negation, (\+)], yes, Goal, [Cin, Cin]|_],
    !,
    link_ho_goal(Goal, [Cin, Cin]),
    append(InCtrVars, [Cin, Cin], OutCtrVars)
  ;
    % Weak unification
    Functor == weak_unify_a3,
    Args = [_, _, _Lambda, Cin, Cout|_],
    !,
    append(InCtrVars, [Cin, Cout], OutCtrVars)
  ;
    % Higher order call
    Functor == ho_call,
    Args = [_HOList, _Compiled, _Goal, ConstraintsVars, _GoalDegreeVars|_TermArgs],
    !,
      (nonvar(ConstraintsVars)
      -> append(InCtrVars, ConstraintsVars, OutCtrVars)
      ;  OutCtrVars = InCtrVars)
  ;
    % Hypothetical operator
    flags:get_bpl_flag(hypothetical(true)),
    Term = '=>'(_HypoRules, _Goal),
    !,
    OutCtrVars = InCtrVars
  ;
    % Predefined predicate (built-in) without constraints
    flags:get_bpl_flag(disjunction(true)),
    Term = '$builtin'(_Builtin, [Cin, Cout]),
    !,
    append(InCtrVars, [Cin, Cout], OutCtrVars)
  ;
    % Quoted term, or existential operator
    ( utilities:builtin(Term) ;
      utilities:is_quoted(Functor, '\'') ;
      Functor == '^' ),
    !,
    OutCtrVars = InCtrVars
  ;
    % Defined predicates
    % Extract the constraint variables from the scanned term
    Ctrs = [Cin, Cout],
    (flags:get_bpl_flag(degree_accumulator(false)) 
     ->
      append(_, [Cin, Cout, _DegreeVar], Args)
     ;
      append(_, [Cin, Cout, _DegreeVarIn, _DegreeVarOut], Args)
    ),
    % Add the constraint variables to the output data
    append(InCtrVars, Ctrs, OutCtrVars)
  ),
  !.

extract_constraint_variables_scanner(Term, Term, InData, InData).
  % Term can be ignored because it's a variable or a number


%% get_variables_scanner(+Term, -Result, +InData, -OutData)
%
%     Scanner that can be used with process_term/6 to count the
%     occurrences of each variable in a term and save them in OutData as
%     a list containing [Term, Count] parings. This scanner always
%     unifies Result with Term. Note that, in this context, "variable"
%     means "atom starting with an uppercase letter or an underscore
%     character".
%
%     @see process_term/6
%

get_variables_scanner(Term, Term, CurrentVars, NewVars) :-
  atom(Term),
  % Term is an atom; if its first char is uppercase, then it represents
  % a variable (variables starting with an underscore character are ignored)
  !,
  (utilities:atom_is_variable(Term) ->
    % Term represents a variable, so it must be added to
    % the list of current variables
    (member([Term, Count], CurrentVars) ->
      % This variable has already been found before
      delete(CurrentVars, [Term, Count], CurrentVarsWithoutTerm),
      NewCount is Count + 1,
      NewVars = [[Term, NewCount]|CurrentVarsWithoutTerm]
    ;
      % This is the first ocurrence of this variable
      NewVars = [[Term, 1]|CurrentVars]
    )
  ;
    % Term doesn't represent a variable
    NewVars = CurrentVars
  ).

get_variables_scanner(Term, Term, Vars, Vars).
  % Term can be ignored because it isn't an atom


%% replace_variable_calls_scanner(+Term, -Result, +InData, -OutData)
%
%     Scanner that can be used with process_term/6 to count the
%     occurrences of each variable in a term and save them in OutData as
%     a list containing [Term, Count] parings. This scanner always
%     unifies Result with Term. Note that, in this context, "variable"
%     means "atom starting with an uppercase letter or an underscore
%     character".
%
%     @see process_term/6
%

replace_variable_calls_scanner(Term, NewTerm, Data, Data) :-
  utilities:atom_is_variable(Term),
  !,
  % Term represents a variable, so it must be added to
  % the list of current variables
  NewTerm = bpl_call(Term).

replace_variable_calls_scanner(Term, Term, Data, Data).
  % Term can be ignored because it isn't an atom


%% control_predicate_tester(+Term)
%
%     Tester that can be used with process_term/6 to scan a clause or
%     goal. This tester fails only when Term is a compound term with a
%     functor different from ,/2, ;/2, ->/2, *->/2 and =>/2 (this last 
%     one when hypothetical reasoning is enabled).
%
%     @see process_term/6
%

%% is_body_connective(+Term)
%
%     Helper predicate which is true if Term is a body connective 
%     (',', ';', '->' ...)

control_predicate_tester(Term) :-
  compound(Term),
  !,
  is_body_connective(Term).

control_predicate_tester(_Term).
  % Tester always succeeds for non-compound terms


is_body_connective(Term) :-
  functor(Term, Functor, Arity),
  (Functor == ',' ; 
   Functor == ';' ; 
   Functor == '->' ; 
   Functor == '*->' ;
   (Functor == '=>', Arity == 2, flags:get_bpl_flag(hypothetical(true)))
   ).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for hypothetical reasoning
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% hypothetical_term_scanner(+Term, -Result, +[HypoData, InRules], -[HypoData, OutRules])
%
%     Scanner that can be used with process_term/6 to add 
%     hypothetical control data to each call to a 
%     user-defined predicate, and get both the new constraint
%     variables in OutData. For every Term that is a compound term
%     with a non-predefined functor, Result will be unified with a 
%     new term with the same functor as Term and:
%
%      * Arguments: the ones in the original term plus .
%
%     HypoData = hypo(head, HypoTerm)
%              | hypo(body, RuleVars, CD)
%              | hypo(query, SharedVars, CD)
%
%     For example, given:
%         Term   = 'prog_pred(X, 1, DegreeVar)',
%         InData = hypo(SharedVs, RI, RD, CD)
%     this predicate will return 
%         Result  = 'prog_pred(X, 1, Cin, Cout, DegreeVar)', and
%         OutData = InData ++ [Cin, Cout].
%     If constraints are passed due to a2 or a3, then:
%         Result  = 'prog_pred(X, 1, Cin, Cout, DegIn, DegOut)', and
%         OutData = InData ++ [Cin, Cout].
%
%     If Term is a call to a predefined predicate (for example,
%     'append([1], X, [1, 2])'), Result and OutData will be unified
%     with Term and InData, respectively.
%
%     @see process_term/6
%

hypothetical_term_scanner(Term, NewTerm, [HypoData, InRules], [HypoData, OutRules]) :-
  nonvar(Term),
  not(number(Term)),
  !,
  Term =.. [Functor|Args],
  (
    % A variable, so it can be ignored
    utilities:atom_is_variable(Functor),
    !,
    NewTerm = Term,
    OutRules = InRules
  ;
    % Higher order predicate with run-time compilation
    Functor == ho_call,
    Args = [[HOPredicate, _]|_],
    ( HOPredicate == bpl_apply
    ; HOPredicate == bpl_maplist),
    !,
    NewTerm = Term,
    OutRules = InRules
  ;
    % Higher order predicate
    Functor == ho_call,
    !,
    (Args = [_, yes|_] % Add hypothetical term only if Term is already compiled by add_degree_variables
     -> ho_hypothetical_term_scanner(Args, NewTerm, [HypoData, InRules], [HypoData, OutRules]) % WARNING: add_hypothetical_term?
     ;  NewTerm = Term,
        OutRules = InRules )
  ;
    % Call to a predefined predicate or a quoted term,
    % so it can be ignored
    (utilities:builtin(Term) ; 
     utilities:is_quoted(Functor, '\'') ;
     Functor == '^'),
    !,
    NewTerm = Term,
    OutRules = InRules
  ;
    % Embedded implication
    % Term = '=>'('_$'(Rules), Goal),
    Term = '=>'(Left, Goal),
    !,
    (Left = '_$'(Rules) ; Left = Rules), % This is because the constraint variables scanner might not been applied
    !,
    (HypoData = hypo(body, RuleVars, CD)
     ;
     HypoData = hypo(query, _SharedVars, CD),
     RuleVars = []),
    NewTerm = '=>'(IdsVs, NewGoal, DI, CD),
    NewCD = [DI|CD],
    translate_hypo_rules(Rules, RuleVars, IdsVs, NewCD, NewRules1),
    ConclusionHypoData = hypo(body, _Free, NewCD),
    add_hypothetical_term(Goal, NewGoal, ConclusionHypoData, NewRules2),
    append([InRules, NewRules1, NewRules2], OutRules)
  ;
    % Call to defined predicate or rule head
    % Adds the hypothetical term to the scanned term
    % HypoData = hypo(head, CD) 
    %          | hypo(body, RuleVars, CD)
    atom_to_hypothetical_atom(Term, NewTerm, HypoData),
    OutRules = InRules
  ).

hypothetical_term_scanner(Term, Term, [HypoData, InRules], [HypoData, InRules]).
  % Term can be ignored because it's a variable or a number


%% ho_hypothetical_term_scanner(+Args, -NewTerm, [+HypoData, +InRules], [-HypoData, -OutRules])
%
%     Scanner that can be used with process_term/6 to add
%     hypothetical control data to each higher-order call to a
%     user-defined predicate.

ho_hypothetical_term_scanner([HOList|Args], NewTerm, [HypoData, InRules], [HypoData, OutRules]) :-
  % Negation, with the form eval_negation(Negation, Goal), Negation \in {\+, not}
  ( HOList = [Functor],
    ( Functor == 'bpl_call'
    ; higher_order_predicate(Functor, [goal])       % once/1...
    ; higher_order_predicate(Functor, [goal, goal]) % forall/2 
    ; higher_order_predicate(Functor, [goal, nongoal, goal]) % catch/3 
    ),
    NewHOList = HOList
  ; HOList = [Functor, _NonGoal],
    ( Functor == 'eval_negation'
    ; higher_order_predicate(Functor, [goal, nongoal]) 
    ),
    NewHOList = HOList
  ; HOList = [Functor, _Template, _Bag],            % findall/3
    higher_order_predicate(Functor, [nongoal, goal, nongoal]),
    NewHOList = HOList
  ; HOList = [Functor, Template, Bag, Bindings],    % bagof/3, setof/3
    higher_order_predicate(Functor, [nongoal, goal, nongoal])
  ),
  !,
  Args = [Compiled, Goal|CtrDegreeVars],
  add_hypothetical_term(Goal, NewGoal, HypoData, NewHypoRules),
  ((Functor == bagof ; Functor == setof)
   -> term_variables(Goal, GoalVars),
      term_variables(NewGoal, NewGoalVars),
      utilities:var_subtract(NewGoalVars, GoalVars, HypoBindings),
      append(HypoBindings, Bindings, NewBindings),
      NewHOList = [Functor, Template, Bag, NewBindings]
   ;  NewHOList = HOList ),
  append(InRules, NewHypoRules, OutRules),
  NewTerm =.. [ho_call, NewHOList, Compiled, NewGoal|CtrDegreeVars].


%% translate_hypo_rules(+Rules, +AllVariables, +Pairs RuleId-SharedVs, +Current_Database_Context, -TranslatedRules) 
%
%     Translates the conjunction of hypothetical rules in the antecedent
%     of an embedded implication.
%
%     Each rule is translated as any normal rule, but including in the
%     hypothetical data the shared variables (SharedVs) with the rule R
%     in which it occurs, the rule context of R (CD), and a new rule 
%     context (_CD) for the translated rules:
%
%       hypo(SharedVs, RI, CD, _CD)
%

translate_hypo_rules((Rule/\Rules), Vs, [IdVs|IdsVs], CD, [NewRule|NewRules]) :-
  !,
  translate_hypo_rules(Rule, Vs, [IdVs], CD, [NewRule]),
  translate_hypo_rules(Rules, Vs, IdsVs, CD, NewRules).

translate_hypo_rules(Rule, AllVs, [RI-SharedVs], CD, NewRules) :-
  parsed_term_variables(Rule, RuleVs),
  intersect_eq(AllVs, RuleVs, SharedVs),
  translate_rule(Rule, _RuleWeight, hypo(SharedVs, RI, CD, _CD), [0,0], NewRules). %WARNING: Should pass Line and Column data instead of [0,0].


%% init_hypothetical_term(-HypoTerm)
%
%     Prepares a hypothetical term for the user program rule being 
%     parsed.
%
%     HypoTerm is of the form hypo(SharedVs, RI, RD, CD), where:
%
%       SharedVs gets []
%       RD gets the user program database context ([])
%       RI gets its value when translating the rule
%

init_hypothetical_term(hypo([], _RI, [], _CD)) :-
  flags:get_bpl_flag(hypothetical(true)),
  !.

init_hypothetical_term(_).


%% atom_to_hypothetical_atom(+Atom, -Result, ?HypoData)
%
%     Adds the hypothetical term hypo(SharedVars, RI, RD, CD) to 
%     either the head, or predicate call at the body, or
%     rule in an antecedent of an embedded implication, where
%
% HypoData = hypo(head, HypoTerm)
%          | hypo(body, RuleVars, CD)
%          | hypo(query, SharedVars, CD)
%
%     For a head, the hypothetical term is given, 
%     For a call, it is constructed from the given CD

%% add_hypothetical_term_to_atom(+Atom, -Result, ?HypotheticalTerm)
%
%     Helper predicate that adds the hypothetical term to Atom

atom_to_hypothetical_atom(Atom, Result, hypo(head, HypotheticalTerm)) :-
  !,
  add_hypothetical_term_to_atom(Atom, Result, HypotheticalTerm).
  
atom_to_hypothetical_atom(Atom, Result, hypo(body, _RuleVars, CD)) :-
  !,
  HypotheticalTerm = hypo(_Free, _RI, _RD, CD),
  add_hypothetical_term_to_atom(Atom, Result, HypotheticalTerm).

atom_to_hypothetical_atom(Atom, Result, hypo(query, SharedVars, CD)) :-
  !,
  HypotheticalTerm = hypo(SharedVars, _RI, _RD, CD),
  add_hypothetical_term_to_atom(Atom, Result, HypotheticalTerm).


add_hypothetical_term_to_atom(Atom, Result, HypotheticalTerm) :-
  Atom =.. [Functor|Vars],
  length(Vars, ExpandedArityMinus1),
  ExpandedArity is ExpandedArityMinus1 + 1,
  translator:actual_rule_arity(ExpandedArity, ActualRuleArity),
  CtrsDegreesVarsNumber is ExpandedArityMinus1 - ActualRuleArity,
  length(CtrsDegreesVars, CtrsDegreesVarsNumber),
  append(HeadVs, CtrsDegreesVars, Vars),
  append(HeadVs, [HypotheticalTerm|CtrsDegreesVars], NewVars),
  Result =.. [Functor|NewVars].



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates for higher-order applications
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% higher_order_predicate(?Name, ?Arguments)
%
%     Succeeds if Name is the name of a higher-order predicate
%     supported by Bousi-Prolog, and Arguments is a list containing
%     the description of its arguments. Each item of Arguments can be
%     either 'goal', 'nongoal' or 'atom' (just for assert/1 and
%     retract/1).
%     Weak and crisp negation are also supported but are not 
%     included in this list.
%

higher_order_predicate(assert,  [atom]).
higher_order_predicate(retract, [atom]).
higher_order_predicate(call,    [goal]).
higher_order_predicate(once,    [goal]).
higher_order_predicate(ignore,  [goal]).
higher_order_predicate(time,    [goal]).
higher_order_predicate(apply,   [nongoal, nongoal]).
higher_order_predicate(maplist, [nongoal, nongoal]).
higher_order_predicate(forall,  [goal, goal]).
higher_order_predicate(findall, [nongoal, goal, nongoal]).
higher_order_predicate(bagof,   [nongoal, goal, nongoal]).
higher_order_predicate(setof,   [nongoal, goal, nongoal]).
higher_order_predicate(catch,   [goal, nongoal, goal]).
higher_order_predicate(phrase,  [goal, nongoal, nongoal]).
higher_order_predicate(phrase,  [goal, nongoal]).
higher_order_predicate(truth_degree, [goal, nongoal]).


%% is_total_goal(+Goal)
%
%     Succeeds if Goal is total, i.e., it contains no variables
%     at positions in which non-variables are expected.
%

is_total_goal(Goal) :-
  (  var(Goal)
   ; number(Goal)
   ; is_list(Goal) 
   ; atom_is_variable(Goal)),
  !,
  fail.

is_total_goal(Goal) :-
  ( Goal = (Goal1, Goal2)
  ; Goal = (Goal1; Goal2) ),
  !,
  is_total_goal(Goal1),
  is_total_goal(Goal2).

is_total_goal(Goal1 *-> Goal2) :-
  !,
  is_total_goal(Goal1 -> Goal2).

is_total_goal(Goal1 -> Goal2) :-
  !,
  is_total_goal(Goal1),
  is_total_goal(Goal2).

is_total_goal('=>'(Rules, Goal)) :-
  flags:get_bpl_flag(hypothetical(true)),
  !,
  is_total_hypo_rules(Rules),
  is_total_goal(Goal).

is_total_goal(Goal) :-
  compound(Goal),
  Goal =.. [Functor|Arguments],
  (higher_order_predicate(Functor, Pattern)
   ; 
   (Functor == not ; Functor == (\+)), 
   Pattern = [goal]),
  !,
  filter_goal_arguments(Arguments, Pattern, GoalArguments),
  maplist(is_total_goal, GoalArguments).

is_total_goal(_Goal) :-
  !.


%% is_total_rule(+Rule)
%
%     Succeeds if Rule is total, i.e., its body is total and
%     its head is not a variable.
%

is_total_rule(Rule) :-
  atom_is_variable(Rule),
  !,
  fail.

is_total_rule((Head :- _Body)) :-
  atom_is_variable(Head),
  !,
  fail.

is_total_rule((_TotalHead :- Body)) :-
  (is_total_goal(Body)
   -> !
   ;  !, fail).

is_total_rule(_TotalHead).


%% is_total_hypo_rules(+Rules)
%
%     Succeeds if all Rules in a hypothetical assumption
%     are total.
%

is_total_hypo_rules(Rules) :-
  (var(Rules) ; atom_is_variable(Rules)),
  !,
  fail.

is_total_hypo_rules(Rule/\MoreRules) :-
  !,
  is_total_rule(Rule),
  is_total_hypo_rules(MoreRules).

is_total_hypo_rules(Rule) :-
  is_total_rule(Rule).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Miscellaneous predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% parsed_term_variables(+Term, -TermVars)
%
%     Get variables from a parsed term. Each variable in a parsed
%     term is an atom starting with a capital letter or an 
%     underscore. It is only used when hypothetical reasoning is 
%     enabled. Otherwise, it returns an empty list.
%

parsed_term_variables(_Term, []) :-
  flags:get_bpl_flag(hypothetical(false)),
  !.
  
parsed_term_variables(Term, TermVars) :-
  parsed_term_variables_aux(Term, DuplicatedTermVars, []),
  sort(DuplicatedTermVars, TermVars).


parsed_term_variables_aux(Term) -->
  { var(Term) ; number(Term) },
  [],
  !.

parsed_term_variables_aux([]) -->
  [],
  !.

parsed_term_variables_aux([X|Xs]) -->
  parsed_term_variables_aux(X),
  parsed_term_variables_aux(Xs),
  !.

parsed_term_variables_aux(Term) -->
  [Term],
  { \+ compound(Term),
    atom_is_variable(Term),
    ! }.

parsed_term_variables_aux(Term) -->
  { compound(Term),
    Term =.. [_Functor|Arguments],
    ! },
  parsed_term_variables_aux(Arguments).

parsed_term_variables_aux(_Term) -->
  [].


%% valid_parsed_head(+Term)
%
%     True if Term is a valid parsed rule head, i.e., it is neither
%     a variable nor a number
%
valid_parsed_head(Term) :-
    \+ utilities:atom_is_variable(Term),
    \+ number(Term).


%% term_constraint_variables(+Term, -Vars)
%
%    Accesses the constraint variables 
%    of a term (head or goal)
%

term_constraint_variables(_Term, [_, _]) :-
  flags:get_bpl_flag(weak_unification(a1)),
  !.
  
term_constraint_variables(Term, [Cin, Cout]) :-
  flags:get_bpl_flag(degree_accumulator(false)),
  !,
  (atom_is_variable(Term)
   -> true
   ;  Term =.. [_|Args],
      append(_, [Cin, Cout, _Degree], Args)).
  
term_constraint_variables(Term, [Cin, Cout]) :-
  (atom_is_variable(Term)
   -> true
   ;  Term =.. [_|Args],
      append(_, [Cin, Cout, _DegreeIn, _DegreeOut], Args)).
  
  
%% term_degree_vars(+Term, -Vars)
%
%    Accesses the approximation degree variables 
%    of a term (head or goal)
%

term_degree_vars(Term, [Degree]) :-
  flags:get_bpl_flag(degree_accumulator(false)),
  !,
  (atom_is_variable(Term)
   -> true
   ;  Term =.. [_|Args],
      append(_, [Degree], Args)).
  
term_degree_vars(Term, [DegreeIn, DegreeOut]) :-
  (atom_is_variable(Term)
   -> true
   ;  Term =.. [_|Args],
      append(_, [DegreeIn, DegreeOut], Args) ).
  
  
%% declare_custom_operator(+Precedence, +Type, +Names)
%
%     Adds each of the operators in Names list to the custom operator
%     list, using specified type and precedence.
%

declare_custom_operator(_, _, []).

declare_custom_operator(Precedence, Type, [Name|MoreOps]) :-
  custom_operators(CurrentOpList),
  retract(custom_operators(CurrentOpList)),
  assert(custom_operators([[Precedence, Type, Name]|CurrentOpList])),
  declare_custom_operator(Precedence, Type, MoreOps).


%% operator_type(?Precedence, ?Type, +Name)
%
%     Succeeds if Name is currently defined as a predefined or custom
%     operator with specified type and precedence.
%

operator_type(Precedence, Type, Name) :-
  % Checks if the operator is a predefined one
  evaluator:current_op(Precedence, Type, Name),
  Name \== '.'. % SWI-Prolog 7.x add the dot as an infix operator for dicts.

operator_type(Precedence, Type, Name) :-
  % Checks if the operator is a custom operator declared
  % by the user with the op/3 directive
  custom_operators(CustomOpList),
  member([Precedence, Type, Name], CustomOpList).


%% bind_term(+Variables, +Term, -BoundTerm)
%
%     Binds Term with all the variables in the Variables list using
%     the existencial quantification operator (^) and returns the
%     result in BoundTerm. For example, given Variables = [A, B] and
%     Term = 'pred(A, B, C)', this predicate will return BoundTerm =
%     'A ^ B ^ pred(A, B, C)'.
%

bind_term([], Term, Term).

bind_term([Var|MoreVars], Term, BoundTerm) :-
  BoundTerm = Var ^ ContTerm,
  bind_term(MoreVars, Term, ContTerm).


%% unbind_term(+BoundTerm, -UnboundTerm, -Variables)
%
%     Unbinds Term.
%

unbind_term(X^Goal, NewGoal, [X|Bindings]) :-
  !,
  unbind_term(Goal, NewGoal, Bindings).

unbind_term(Goal, Goal, []).
  

%% reset_parser
%
%     Resets custom operators, messages and linguistic terms.
%

reset_parser :-
  custom_operators(OldOperators),
  retract(custom_operators(OldOperators)),
  assert(custom_operators([])),
  messages(OldMessages),
  retract(messages(OldMessages)),
  assert(messages([])),
  linguistic_terms(OldLingTerms),
  retract(linguistic_terms(OldLingTerms)),
  assert(linguistic_terms([])),
  reset_ri. % Rule identifier for hypothetical reasoning


%% filter_goal_arguments(+Arguments, +Pattern, -GoalArguments)
filter_goal_arguments([], [], []).

filter_goal_arguments([Argument|Arguments], [goal|Pattern], [Argument|GoalArguments]) :-
  !,
  filter_goal_arguments(Arguments, Pattern, GoalArguments).

filter_goal_arguments([_Argument|Arguments], [_|Pattern], GoalArguments) :-
  filter_goal_arguments(Arguments, Pattern, GoalArguments).



%% BEGIN FROM SWI-Prolog hprolog.pl

%%  memberchk_eq(+Val, +List)
%
%  Deterministic check of membership using == rather than
%  unification.

memberchk_eq(X, [Y|Ys]) :-
   (   X == Y
   ->  true
   ;   memberchk_eq(X, Ys)
   ).


%%  intersect_eq(+List1, +List2, -Intersection)
%
%  Determine the intersection of two lists without unifying values.

intersect_eq([], _, []).

intersect_eq([X|Xs], Ys, L) :-
  (   memberchk_eq(X, Ys)
  ->  L = [X|T],
      intersect_eq(Xs, Ys, T)
  ;   intersect_eq(Xs, Ys, L)
  ).
  
%% END FROM SWI-Prolog hprolog.pl



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% program_prefix(?Name)                                     is dynamic
%
%     Dynamic predicate which stores the prefix corresponding to the
%     main BPL source code file (if a program is being parsed), or to
%     the currently loaded Bousi-Prolog program (if a query is being
%     parsed).
%

:- dynamic program_prefix/1.

program_prefix('none').


%% bpl_filename(?Filename)                                   is dynamic
%
%     Dynamic predicate which stores the name of the BPL source code
%     file that's being parsed.
%

:- dynamic bpl_filename/1.

bpl_filename('').


%% messages(?Messages)                                       is dynamic
%
%     Dynamic predicate which stores all warnings and errors generated
%     during lexical, syntactic and semantic analysis. Each warning and
%     error must be a list with five items: file name, line number,
%     column number, message text, and message type ('warning' or
%     'error').
%

:- dynamic messages/1.

messages([]).


%% custom_operators(?OpList)                                 is dynamic
%
%     Dynamic predicate which stores custom operators declared in
%     source code files using the op/3 Prolog directive. Each operator
%     is defined by a list with three items: precedence, type, and
%     name.
%

:- dynamic custom_operators/1.

custom_operators([]).


%% linguistic_terms(?LingTerms)                              is dynamic
%
%     Dynamic predicate which stores all linguistic terms built with
%     the '#' operator found in source code (such as 'age#30' or
%     'very#tall'). Each linguistic term is defined using a list with
%     three items: the first two can be 'domain' followed by the name
%     of an existing domain or 'subset' followed by the name of a
%     fuzzy subset, while the last one must be the definition of the
%     subset, as required by ext_translate_fuzzysets/5.
%

:- dynamic linguistic_terms/1.

linguistic_terms([]).


%% ri(?Id)                                                   is dynamic
%
%     Dynamic predicate which stores the identifier for the next
%     hypothetical rule identifier

:- dynamic(ri/1). 

get_ri(RI) :-
  retract(ri(RI)),
  RI1 is RI+1,
  assertz(ri(RI1)),
  !.
  
reset_ri :-
  retractall(ri(_)),
  assertz(ri(0)).

:- reset_ri. 


